from __future__ import annotations

import logging
from collections.abc import Container, Iterable, Sequence
from dataclasses import dataclass, field
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Self,
    TypedDict,
    override,
)

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.engine import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import Mapped, foreign, load_only, mapped_column, relationship

from ai.backend.common import msgpack
from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.domain.types import DomainData
from ai.backend.manager.defs import RESERVED_DOTFILES
from ai.backend.manager.models.base import (
    Base,
    ResourceSlotColumn,
    SlugType,
    VFolderHostPermissionColumn,
)
from ai.backend.manager.models.rbac import (
    AbstractPermissionContext,
    AbstractPermissionContextBuilder,
    DomainScope,
    ProjectScope,
    RBACModel,
    ScopeType,
    UserScope,
    get_predefined_roles_in_scope,
    required_permission,
)
from ai.backend.manager.models.rbac.context import ClientContext
from ai.backend.manager.models.rbac.permission_defs import DomainPermission

if TYPE_CHECKING:
    from ai.backend.manager.models.group import GroupRow
    from ai.backend.manager.models.network import NetworkRow
    from ai.backend.manager.models.scaling_group import ScalingGroupForDomainRow
    from ai.backend.manager.models.session import SessionRow
    from ai.backend.manager.models.user import UserRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


__all__: Sequence[str] = (
    "MAXIMUM_DOTFILE_SIZE",
    "DomainDotfile",
    "DomainRow",
    "domains",
    "query_domain_dotfiles",
    "verify_dotfile_name",
)

MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB


def row_to_data(row: DomainRow | Row[Any]) -> DomainData:
    return DomainData(
        name=row.name,
        description=row.description,
        is_active=row.is_active,
        created_at=row.created_at,
        modified_at=row.modified_at,
        total_resource_slots=row.total_resource_slots,
        allowed_vfolder_hosts=row.allowed_vfolder_hosts,
        allowed_docker_registries=row.allowed_docker_registries,
        integration_id=row.integration_id,
        dotfiles=row.dotfiles,
    )


def _get_network_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.network import NetworkRow

    return DomainRow.name == foreign(NetworkRow.domain_name)


class DomainRow(Base):  # type: ignore[misc]
    __tablename__ = "domains"

    name: Mapped[str] = mapped_column(
        "name", SlugType(length=64, allow_unicode=True, allow_dot=True), primary_key=True
    )
    description: Mapped[str | None] = mapped_column("description", sa.String(length=512))
    is_active: Mapped[bool] = mapped_column("is_active", sa.Boolean, default=True, nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now(), nullable=False
    )
    modified_at: Mapped[datetime] = mapped_column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
        nullable=False,
    )
    # TODO: separate resource-related fields with new domain resource policy table when needed.
    total_resource_slots: Mapped[ResourceSlot] = mapped_column(
        "total_resource_slots", ResourceSlotColumn(), default=ResourceSlot, nullable=False
    )
    allowed_vfolder_hosts: Mapped[VFolderHostPermissionMap] = mapped_column(
        "allowed_vfolder_hosts",
        VFolderHostPermissionColumn(),
        nullable=False,
        default=VFolderHostPermissionMap,
    )
    allowed_docker_registries: Mapped[list[str]] = mapped_column(
        "allowed_docker_registries", pgsql.ARRAY(sa.String), nullable=False, default=list
    )
    #: Field for synchronization with external services.
    integration_id: Mapped[str | None] = mapped_column("integration_id", sa.String(length=512))
    # dotfiles column, \x90 means empty list in msgpack
    dotfiles: Mapped[bytes] = mapped_column(
        "dotfiles", sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b"\x90"
    )

    sessions: Mapped[list[SessionRow]] = relationship("SessionRow", back_populates="domain")
    users: Mapped[list[UserRow]] = relationship("UserRow", back_populates="domain")
    groups: Mapped[list[GroupRow]] = relationship("GroupRow", back_populates="domain")
    sgroup_for_domains_rows: Mapped[list[ScalingGroupForDomainRow]] = relationship(
        "ScalingGroupForDomainRow",
        back_populates="domain_row",
    )
    networks: Mapped[list[NetworkRow]] = relationship(
        "NetworkRow",
        back_populates="domain_row",
        primaryjoin=_get_network_join_condition,
    )

    def to_data(self) -> DomainData:
        return row_to_data(self)


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use DomainRow class directly for new code.
domains = DomainRow.__table__


@dataclass
class DomainModel(RBACModel[DomainPermission]):
    name: str
    description: str | None
    is_active: bool
    created_at: datetime
    modified_at: datetime

    _total_resource_slots: ResourceSlot
    _allowed_vfolder_hosts: VFolderHostPermissionMap
    _allowed_docker_registries: list[str]
    _integration_id: str | None
    _dotfiles: bytes

    orm_obj: DomainRow
    _permissions: frozenset[DomainPermission] = field(default_factory=frozenset)

    @property
    def permissions(self) -> Container[DomainPermission]:
        return self._permissions

    @property
    @required_permission(DomainPermission.READ_SENSITIVE_ATTRIBUTE)
    def total_resource_slots(self) -> ResourceSlot:
        return self._total_resource_slots

    @property
    @required_permission(DomainPermission.READ_SENSITIVE_ATTRIBUTE)
    def allowed_vfolder_hosts(self) -> VFolderHostPermissionMap:
        return self._allowed_vfolder_hosts

    @property
    @required_permission(DomainPermission.READ_SENSITIVE_ATTRIBUTE)
    def allowed_docker_registries(self) -> list[str]:
        return self._allowed_docker_registries

    @property
    @required_permission(DomainPermission.READ_SENSITIVE_ATTRIBUTE)
    def integration_id(self) -> str | None:
        return self._integration_id

    @property
    @required_permission(DomainPermission.READ_SENSITIVE_ATTRIBUTE)
    def dotfiles(self) -> bytes:
        return self._dotfiles

    @classmethod
    def from_row(cls, row: DomainRow, permissions: Iterable[DomainPermission]) -> Self:
        return cls(
            name=row.name,
            description=row.description,
            is_active=row.is_active,
            created_at=row.created_at,
            modified_at=row.modified_at,
            _total_resource_slots=row.total_resource_slots,
            _allowed_vfolder_hosts=row.allowed_vfolder_hosts,
            _allowed_docker_registries=row.allowed_docker_registries,
            _integration_id=row.integration_id,
            _dotfiles=row.dotfiles,
            _permissions=frozenset(permissions),
            orm_obj=row,
        )


class DomainDotfile(TypedDict):
    data: str
    path: str
    perm: str


async def query_domain_dotfiles(
    conn: SAConnection,
    name: str,
) -> tuple[list[DomainDotfile], int]:
    query = sa.select(DomainRow.dotfiles).where(DomainRow.name == name)
    packed_dotfile = await conn.scalar(query)
    if packed_dotfile is None:
        return [], MAXIMUM_DOTFILE_SIZE
    rows = msgpack.unpackb(packed_dotfile)
    return rows, MAXIMUM_DOTFILE_SIZE - len(packed_dotfile)


def verify_dotfile_name(dotfile: str) -> bool:
    return dotfile not in RESERVED_DOTFILES


ALL_DOMAIN_PERMISSIONS = frozenset([perm for perm in DomainPermission])
OWNER_PERMISSIONS: frozenset[DomainPermission] = ALL_DOMAIN_PERMISSIONS
ADMIN_PERMISSIONS: frozenset[DomainPermission] = ALL_DOMAIN_PERMISSIONS
MONITOR_PERMISSIONS: frozenset[DomainPermission] = frozenset([
    DomainPermission.READ_ATTRIBUTE,
    DomainPermission.UPDATE_ATTRIBUTE,
])
PRIVILEGED_MEMBER_PERMISSIONS: frozenset[DomainPermission] = frozenset([
    DomainPermission.READ_ATTRIBUTE
])
MEMBER_PERMISSIONS: frozenset[DomainPermission] = frozenset([DomainPermission.READ_ATTRIBUTE])

type WhereClauseType = sa.sql.expression.BinaryExpression[Any] | sa.sql.expression.BooleanClauseList


@dataclass
class DomainPermissionContext(AbstractPermissionContext[DomainPermission, DomainRow, str]):
    @property
    def query_condition(self) -> WhereClauseType | None:
        cond: WhereClauseType | None = None

        def _OR_coalesce(
            base_cond: WhereClauseType | None,
            _cond: sa.sql.expression.BinaryExpression[Any],
        ) -> WhereClauseType:
            return base_cond | _cond if base_cond is not None else _cond

        if self.object_id_to_additional_permission_map:
            cond = _OR_coalesce(
                cond, DomainRow.name.in_(self.object_id_to_additional_permission_map.keys())
            )
        if self.object_id_to_overriding_permission_map:
            cond = _OR_coalesce(
                cond, DomainRow.name.in_(self.object_id_to_overriding_permission_map.keys())
            )
        return cond

    async def build_query(self) -> sa.sql.Select[Any] | None:
        cond = self.query_condition
        if cond is None:
            return None
        return sa.select(DomainRow).where(cond)

    async def calculate_final_permission(self, rbac_obj: DomainRow) -> frozenset[DomainPermission]:
        domain_row = rbac_obj
        domain_name = domain_row.name
        permissions: frozenset[DomainPermission] = frozenset()

        if (
            overriding_perm := self.object_id_to_overriding_permission_map.get(domain_name)
        ) is not None:
            permissions = overriding_perm
        else:
            permissions |= self.object_id_to_additional_permission_map.get(domain_name, set())
        return permissions


class DomainPermissionContextBuilder(
    AbstractPermissionContextBuilder[DomainPermission, DomainPermissionContext]
):
    db_session: SASession

    def __init__(self, db_session: SASession) -> None:
        self.db_session = db_session

    @override
    async def calculate_permission(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[DomainPermission]:
        roles = await get_predefined_roles_in_scope(ctx, target_scope, self.db_session)
        return await self._calculate_permission_by_predefined_roles(roles)

    @override
    async def build_ctx_in_system_scope(
        self,
        ctx: ClientContext,
    ) -> DomainPermissionContext:
        perm_ctx = DomainPermissionContext()
        _domain_query_stmt = sa.select(DomainRow).options(load_only(DomainRow.name))
        for row in await self.db_session.scalars(_domain_query_stmt):
            to_be_merged = await self.build_ctx_in_domain_scope(ctx, DomainScope(row.name))
            perm_ctx.merge(to_be_merged)
        return perm_ctx

    @override
    async def build_ctx_in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> DomainPermissionContext:
        permissions = await self.calculate_permission(ctx, scope)
        return DomainPermissionContext(
            object_id_to_additional_permission_map={scope.domain_name: permissions}
        )

    @override
    async def build_ctx_in_project_scope(
        self, ctx: ClientContext, scope: ProjectScope
    ) -> DomainPermissionContext:
        return DomainPermissionContext()

    @override
    async def build_ctx_in_user_scope(
        self, ctx: ClientContext, scope: UserScope
    ) -> DomainPermissionContext:
        return DomainPermissionContext()

    @override
    @classmethod
    async def _permission_for_owner(
        cls,
    ) -> frozenset[DomainPermission]:
        return OWNER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_admin(
        cls,
    ) -> frozenset[DomainPermission]:
        return ADMIN_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_monitor(
        cls,
    ) -> frozenset[DomainPermission]:
        return MONITOR_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_privileged_member(
        cls,
    ) -> frozenset[DomainPermission]:
        return PRIVILEGED_MEMBER_PERMISSIONS

    @override
    @classmethod
    async def _permission_for_member(
        cls,
    ) -> frozenset[DomainPermission]:
        return MEMBER_PERMISSIONS


async def get_permission_ctx(
    target_scope: ScopeType,
    requested_permission: DomainPermission,
    *,
    ctx: ClientContext,
    db_session: SASession,
) -> DomainPermissionContext:
    builder = DomainPermissionContextBuilder(db_session)
    return await builder.build(ctx, target_scope, requested_permission)


async def get_domains(
    target_scope: ScopeType,
    requested_permission: DomainPermission,
    domain_names: Iterable[str] | None = None,
    *,
    ctx: ClientContext,
    db_session: SASession,
) -> list[DomainModel]:
    ret: list[DomainModel] = []
    permission_ctx = await get_permission_ctx(
        target_scope, requested_permission, ctx=ctx, db_session=db_session
    )
    cond = permission_ctx.query_condition
    if cond is None:
        return ret
    query_stmt = sa.select(DomainRow).where(cond)
    if domain_names is not None:
        query_stmt = query_stmt.where(DomainRow.name.in_(domain_names))
    async for row in await db_session.stream_scalars(query_stmt):
        permissions = await permission_ctx.calculate_final_permission(row)
        ret.append(DomainModel.from_row(row, permissions))
    return ret



from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence
from datetime import datetime
from typing import Any
from uuid import UUID

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.errors.resource import InvalidUserUpdateMode
from ai.backend.manager.models.group.row import GroupRow
from ai.backend.manager.models.kernel import KernelRow
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.querier import BatchQuerier
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.group.db_source import GroupDBSource
from ai.backend.manager.repositories.group.types import (
    DomainProjectSearchScope,
    GroupSearchResult,
    UserProjectSearchScope,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


group_repository_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.GROUP_REPOSITORY)),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class GroupRepository:
    _db_source: GroupDBSource
    _config_provider: ManagerConfigProvider
    _valkey_stat_client: ValkeyStatClient
    _storage_manager: StorageSessionManager

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        config_provider: ManagerConfigProvider,
        valkey_stat_client: ValkeyStatClient,
        storage_manager: StorageSessionManager,
    ) -> None:
        self._db_source = GroupDBSource(db)
        self._config_provider = config_provider
        self._valkey_stat_client = valkey_stat_client
        self._storage_manager = storage_manager

    @group_repository_resilience.apply()
    async def create(self, creator: Creator[GroupRow]) -> GroupData:
        """Create a new group."""
        return await self._db_source.create(creator)

    @group_repository_resilience.apply()
    async def modify_validated(
        self,
        updater: Updater[GroupRow],
        user_update_mode: str | None = None,
        user_uuids: list[uuid.UUID] | None = None,
    ) -> GroupData | None:
        """Modify a group with validation."""
        if user_update_mode not in (None, "add", "remove"):
            raise InvalidUserUpdateMode("invalid user_update_mode")
        return await self._db_source.modify_validated(updater, user_update_mode, user_uuids)

    @group_repository_resilience.apply()
    async def mark_inactive(self, group_id: uuid.UUID) -> None:
        """Mark a group as inactive (soft delete)."""
        await self._db_source.mark_inactive(group_id)

    @group_repository_resilience.apply()
    async def get_container_stats_for_period(
        self,
        start_date: datetime,
        end_date: datetime,
        group_ids: Sequence[UUID] | None = None,
    ) -> list[dict[str, Any]]:
        """Get container statistics for groups within a time period."""
        return await self._db_source.get_container_stats_for_period(
            start_date,
            end_date,
            self._valkey_stat_client,
            self._config_provider,
            group_ids,
        )

    @group_repository_resilience.apply()
    async def fetch_project_resource_usage(
        self,
        start_date: datetime,
        end_date: datetime,
        project_ids: Sequence[UUID] | None = None,
    ) -> list[KernelRow]:
        """Fetch resource usage data for projects."""
        return await self._db_source.fetch_project_resource_usage(start_date, end_date, project_ids)

    @group_repository_resilience.apply()
    async def purge_group(self, group_id: uuid.UUID) -> bool:
        """Completely remove a group and all its associated data."""
        return await self._db_source.purge_group(group_id, self._storage_manager)

    @group_repository_resilience.apply()
    async def get_project(self, project_id: UUID) -> GroupData:
        """Get a single project by UUID.

        Args:
            project_id: UUID of the project.

        Returns:
            GroupData for the project.

        Raises:
            ProjectNotFound: If project does not exist.
        """
        return await self._db_source.get_project(project_id)

    @group_repository_resilience.apply()
    async def search_projects(
        self,
        querier: BatchQuerier,
    ) -> GroupSearchResult:
        """Search all projects (admin only).

        Args:
            querier: BatchQuerier containing conditions, orders, and pagination.

        Returns:
            GroupSearchResult with items, total_count, and pagination flags.
        """
        return await self._db_source.search_projects(querier)

    @group_repository_resilience.apply()
    async def search_projects_by_domain(
        self,
        scope: DomainProjectSearchScope,
        querier: BatchQuerier,
    ) -> GroupSearchResult:
        """Search projects within a domain.

        Args:
            scope: DomainProjectSearchScope defining the domain to search within.
            querier: BatchQuerier containing conditions, orders, and pagination.

        Returns:
            GroupSearchResult with items, total_count, and pagination flags.
        """
        return await self._db_source.search_projects_by_domain(scope, querier)

    @group_repository_resilience.apply()
    async def search_projects_by_user(
        self,
        scope: UserProjectSearchScope,
        querier: BatchQuerier,
    ) -> GroupSearchResult:
        """Search projects a user is member of.

        Args:
            scope: UserProjectSearchScope defining the user to search for.
            querier: BatchQuerier containing conditions, orders, and pagination.

        Returns:
            GroupSearchResult with items, total_count, and pagination flags.
        """
        return await self._db_source.search_projects_by_user(scope, querier)



from __future__ import annotations

import logging
import uuid
from typing import Final, override

from ai.backend.common.contexts.request_id import current_request_id
from ai.backend.common.contexts.user import current_user
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.actions.action import BaseAction, BaseActionTriggerMeta, ProcessResult
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.data.audit_log.types import AuditLogData
from ai.backend.manager.repositories.audit_log import AuditLogCreatorSpec, AuditLogRepository
from ai.backend.manager.repositories.base import Creator

_BLANK_ID: Final[str] = "(unknown)"

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AuditLogMonitor(ActionMonitor):
    _repository: AuditLogRepository

    def __init__(self, repository: AuditLogRepository) -> None:
        self._repository = repository

    async def _generate_log(self, action: BaseAction, result: ProcessResult) -> None:
        user = current_user()
        data = AuditLogData(
            id=uuid.uuid4(),
            action_id=result.meta.action_id,
            entity_type=action.entity_type(),
            operation=action.operation_type(),
            created_at=result.meta.started_at,
            description=result.meta.description,
            status=result.meta.status,
            entity_id=result.meta.entity_id or _BLANK_ID,
            request_id=current_request_id() or _BLANK_ID,
            triggered_by=str(user.user_id) if user else None,
            duration=result.meta.duration,
        )
        creator = Creator(
            spec=AuditLogCreatorSpec(
                action_id=data.action_id,
                entity_type=data.entity_type,
                operation=data.operation,
                created_at=data.created_at,
                description=data.description,
                status=data.status,
                entity_id=data.entity_id,
                request_id=data.request_id,
                triggered_by=data.triggered_by,
                duration=data.duration,
            )
        )
        await self._repository.create(creator)

    @override
    async def prepare(self, action: BaseAction, meta: BaseActionTriggerMeta) -> None:
        pass

    @override
    async def done(self, action: BaseAction, result: ProcessResult) -> None:
        await self._generate_log(action, result)



from .base import (
    BaseAction,
    BaseActionResult,
    BaseActionResultMeta,
    BaseActionTriggerMeta,
    ProcessResult,
    SearchActionResult,
    TAction,
    TActionResult,
)
from .batch import (
    BaseBatchAction,
    BaseBatchActionResult,
)

__all__ = (
    "BaseAction",
    "BaseActionResult",
    "BaseActionResultMeta",
    "BaseActionTriggerMeta",
    "BaseBatchAction",
    "BaseBatchActionResult",
    "ProcessResult",
    "SearchActionResult",
    "TAction",
    "TActionResult",
)



"""Utility functions for converting recorder types to domain types."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TypeVar

from ai.backend.manager.data.session.types import SchedulingResult, SubStepResult
from ai.backend.manager.sokovan.recorder.types import ExecutionRecord, StepRecord, StepStatus

EntityIdT = TypeVar("EntityIdT")


def step_record_to_sub_step_result(step: StepRecord) -> SubStepResult:
    """Convert recorder StepRecord to SubStepResult for history storage."""
    result = (
        SchedulingResult.SUCCESS if step.status == StepStatus.SUCCESS else SchedulingResult.FAILURE
    )
    return SubStepResult(
        step=step.name,
        result=result,
        error_code=step.error_code,
        message=step.detail,
        started_at=step.started_at,
        ended_at=step.ended_at,
    )


def extract_sub_steps_for_entity(
    entity_id: EntityIdT,
    records: Mapping[EntityIdT, ExecutionRecord],
) -> list[SubStepResult]:
    """Extract sub_steps for a specific entity from records mapping.

    Args:
        entity_id: The entity ID (SessionId, DeploymentId, etc.) to extract steps for.
        records: A mapping of entity IDs to their ExecutionRecords.

    Returns:
        A list of SubStepResult objects extracted from all phases and steps.
    """
    record = records.get(entity_id)
    if record is None:
        return []

    sub_steps: list[SubStepResult] = []
    for phase in record.phases:
        for step in phase.steps:
            sub_steps.append(step_record_to_sub_step_result(step))
    return sub_steps



from __future__ import annotations

import abc
import logging
from typing import override

from ai.backend.common.container_registry import ContainerRegistryType
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.container_registry.harbor import (
    AbstractPerProjectRegistryQuotaClient,
    HarborAuthArgs,
    HarborProjectInfo,
    PerProjectHarborQuotaClient,
)
from ai.backend.manager.data.container_registry.types import PerProjectContainerRegistryInfo
from ai.backend.manager.errors.common import GenericBadRequest
from ai.backend.manager.models.rbac import ProjectScope
from ai.backend.manager.repositories.container_registry_quota.repository import (
    PerProjectRegistryQuotaRepository,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AbstractPerProjectContainerRegistryQuotaService(abc.ABC):
    async def create_quota(
        self,
        scope_id: ProjectScope,
        quota: int,
    ) -> None:
        raise NotImplementedError

    async def update_quota(
        self,
        scope_id: ProjectScope,
        quota: int,
    ) -> None:
        raise NotImplementedError

    async def delete_quota(
        self,
        scope_id: ProjectScope,
    ) -> None:
        raise NotImplementedError

    async def read_quota(
        self,
        scope_id: ProjectScope,
    ) -> int:
        raise NotImplementedError


class PerProjectContainerRegistryQuotaClientPool(abc.ABC):
    def make_client(self, type_: ContainerRegistryType) -> AbstractPerProjectRegistryQuotaClient:
        match type_:
            case ContainerRegistryType.HARBOR2:
                return PerProjectHarborQuotaClient()
            case _:
                raise GenericBadRequest(
                    f"{type_} does not support registry quota per project management."
                )


class PerProjectContainerRegistryQuotaService(AbstractPerProjectContainerRegistryQuotaService):
    _repository: PerProjectRegistryQuotaRepository
    _client_pool: PerProjectContainerRegistryQuotaClientPool

    def __init__(
        self,
        repository: PerProjectRegistryQuotaRepository,
        client_pool: PerProjectContainerRegistryQuotaClientPool,
    ) -> None:
        self._repository = repository
        self._client_pool = client_pool

    def _registry_row_to_harbor_project_info(
        self, registry_info: PerProjectContainerRegistryInfo
    ) -> HarborProjectInfo:
        return HarborProjectInfo(
            url=registry_info.url,
            project=registry_info.project,
            ssl_verify=registry_info.ssl_verify,
        )

    @override
    async def create_quota(
        self,
        scope_id: ProjectScope,
        quota: int,
    ) -> None:
        registry_info = await self._repository.fetch_container_registry_row(scope_id)
        client = self._client_pool.make_client(registry_info.type)
        project_info = self._registry_row_to_harbor_project_info(registry_info)
        credential = HarborAuthArgs(
            username=registry_info.username, password=registry_info.password
        )
        await client.create_quota(project_info, quota, credential)

    @override
    async def update_quota(
        self,
        scope_id: ProjectScope,
        quota: int,
    ) -> None:
        registry_info = await self._repository.fetch_container_registry_row(scope_id)
        client = self._client_pool.make_client(registry_info.type)
        project_info = self._registry_row_to_harbor_project_info(registry_info)
        credential = HarborAuthArgs(
            username=registry_info.username, password=registry_info.password
        )
        await client.update_quota(project_info, quota, credential)

    @override
    async def delete_quota(
        self,
        scope_id: ProjectScope,
    ) -> None:
        registry_info = await self._repository.fetch_container_registry_row(scope_id)
        client = self._client_pool.make_client(registry_info.type)
        project_info = self._registry_row_to_harbor_project_info(registry_info)
        credential = HarborAuthArgs(
            username=registry_info.username, password=registry_info.password
        )
        await client.delete_quota(project_info, credential)

    @override
    async def read_quota(self, scope_id: ProjectScope) -> int:
        registry_info = await self._repository.fetch_container_registry_row(scope_id)
        client = self._client_pool.make_client(registry_info.type)
        project_info = self._registry_row_to_harbor_project_info(registry_info)
        credential = HarborAuthArgs(
            username=registry_info.username, password=registry_info.password
        )
        return await client.read_quota(project_info, credential)



"""Export-related exceptions."""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class ExportReportNotFound(BackendAIError, web.HTTPNotFound):
    """Raised when requested export report is not found."""

    error_type = "https://api.backend.ai/probs/export-report-not-found"
    error_title = "Export report not found."

    def __init__(self, report_key: str) -> None:
        super().__init__(extra_msg=f"Unknown report: {report_key}")

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class InvalidExportFieldKeys(BackendAIError, web.HTTPBadRequest):
    """Raised when requested field keys are invalid."""

    error_type = "https://api.backend.ai/probs/invalid-export-field-keys"
    error_title = "Invalid export field keys."

    def __init__(self, invalid_keys: list[str]) -> None:
        super().__init__(extra_msg=f"Invalid field keys: {invalid_keys}")

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class TooManyConcurrentExports(BackendAIError, web.HTTPServiceUnavailable):
    """Raised when concurrent export limit is exceeded."""

    error_type = "https://api.backend.ai/probs/too-many-concurrent-exports"
    error_title = "Too many concurrent exports."

    def __init__(self) -> None:
        super().__init__(extra_msg="Too many concurrent export requests. Please try again later.")

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.UNAVAILABLE,
        )



from __future__ import annotations

import enum
import uuid
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, override

from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.data.permission.types import (
    EntityType,
    OperationType,
    ScopeType,
)
from ai.backend.manager.errors.resource import DataTransformationFailed
from ai.backend.manager.types import OptionalState, PartialModifier, TriState


class ProjectType(enum.StrEnum):
    GENERAL = "general"
    MODEL_STORE = "model-store"

    @classmethod
    def _missing_(cls, value: Any) -> ProjectType | None:
        if not isinstance(value, str):
            raise DataTransformationFailed(
                f"ProjectType value must be a string, got {type(value).__name__}"
            )
        match value.upper():
            case "GENERAL":
                return cls.GENERAL
            case "MODEL_STORE" | "MODEL-STORE":
                return cls.MODEL_STORE
            case _:
                return None


@dataclass
class GroupData:
    id: uuid.UUID = field(compare=False)
    name: str
    description: str | None
    is_active: bool | None
    created_at: datetime | None = field(compare=False)
    modified_at: datetime | None = field(compare=False)
    integration_id: str | None
    domain_name: str
    total_resource_slots: ResourceSlot
    allowed_vfolder_hosts: VFolderHostPermissionMap
    dotfiles: bytes
    resource_policy: str
    type: ProjectType
    container_registry: dict[str, str] | None

    def scope_id(self) -> ScopeId:
        return ScopeId(
            scope_type=ScopeType.PROJECT,
            scope_id=str(self.id),
        )

    def role_name(self) -> str:
        return f"project-{str(self.id)[:8]}-admin"

    def entity_operations(self) -> Mapping[EntityType, Iterable[OperationType]]:
        return {
            entity: OperationType.admin_operations()
            for entity in EntityType.admin_accessible_entity_types_in_project()
        }


@dataclass
class GroupModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    description: TriState[str] = field(
        default_factory=TriState[str].nop,
    )
    is_active: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)
    domain_name: OptionalState[str] = field(
        default_factory=OptionalState[str].nop,
    )
    total_resource_slots: OptionalState[ResourceSlot] = field(
        default_factory=OptionalState[ResourceSlot].nop
    )
    allowed_vfolder_hosts: OptionalState[dict[str, str]] = field(
        default_factory=OptionalState[dict[str, str]].nop
    )
    integration_id: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    resource_policy: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    container_registry: TriState[dict[str, str]] = field(
        default_factory=TriState[dict[str, str]].nop
    )

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.description.update_dict(to_update, "description")
        self.is_active.update_dict(to_update, "is_active")
        self.domain_name.update_dict(to_update, "domain_name")
        self.total_resource_slots.update_dict(to_update, "total_resource_slots")
        self.allowed_vfolder_hosts.update_dict(to_update, "allowed_vfolder_hosts")
        self.integration_id.update_dict(to_update, "integration_id")
        self.resource_policy.update_dict(to_update, "resource_policy")
        self.container_registry.update_dict(to_update, "container_registry")
        return to_update



"""Cluster configuration validation rule."""

from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationContext,
    SessionCreationSpec,
)

from .base import SessionValidatorRule


class ClusterValidationRule(SessionValidatorRule):
    """Validates cluster configuration for multi-container sessions."""

    def name(self) -> str:
        return "cluster_validation"

    def validate(
        self,
        spec: SessionCreationSpec,
        _context: SessionCreationContext,
    ) -> None:
        """Validate cluster configuration and kernel specifications."""
        # Check if kernel_specs exists
        if not spec.kernel_specs:
            raise InvalidAPIParameters("At least one kernel specification is required")

        # Only validate multi-container sessions further
        if spec.cluster_size <= 1:
            return

        kernel_specs_count = len(spec.kernel_specs)

        if kernel_specs_count == 1:
            # Single kernel spec will be replicated - this is valid
            # The main kernel config will be replicated to sub-containers
            pass
        elif kernel_specs_count > 1:
            # Each container should have its own kernel_config
            if kernel_specs_count != spec.cluster_size:
                raise InvalidAPIParameters(
                    "The number of kernel configs differs from the cluster size"
                )



from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.events.types import AbstractBroadcastEvent
from ai.backend.common.types import AgentId
from ai.backend.manager.errors.common import InternalServerError


class PropagatorEventHandler:
    _event_hub: EventHub

    def __init__(self, event_hub: EventHub) -> None:
        self._event_hub = event_hub

    async def propagate_event(
        self,
        _context: None,
        _source: AgentId,
        event: AbstractBroadcastEvent,
    ) -> None:
        # Generate events to propagate (default implementation returns [self])
        individual_events = event.generate_events()
        for individual_event in individual_events:
            await self._event_hub.propagate_event(individual_event)

    async def propagate_event_with_close(
        self,
        _context: None,
        _source: AgentId,
        event: AbstractBroadcastEvent,
    ) -> None:
        # Generate events to propagate (default implementation returns [self])
        individual_events = event.generate_events()
        for individual_event in individual_events:
            await self._event_hub.propagate_event(individual_event)
            # Close each individual event's domain
            domain_id = individual_event.domain_id()
            if domain_id is None:
                raise InternalServerError(
                    f"Event domain ID is None for {individual_event.event_name()}"
                )
            await self._event_hub.close_by_alias(individual_event.event_domain(), domain_id)



"""Export report definitions."""

from .audit_log import AUDIT_LOG_REPORT
from .project import PROJECT_REPORT
from .session import SESSION_REPORT
from .user import USER_REPORT

__all__ = (
    "AUDIT_LOG_REPORT",
    "PROJECT_REPORT",
    "SESSION_REPORT",
    "USER_REPORT",
)



"""
REST API handlers for deployment system.
Provides CRUD endpoints for deployments and revisions.
"""

from __future__ import annotations

from .handler import create_app

__all__ = ("create_app",)



from __future__ import annotations

from dataclasses import dataclass, field
from decimal import Decimal
from typing import Any, override

from ai.backend.common.types import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    ResourceSlot,
    RuntimeVariant,
)
from ai.backend.manager.data.model_serving.modifier import ExtraMount, ImageRef
from ai.backend.manager.models.endpoint import EndpointAutoScalingRuleRow, EndpointRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState, TriState


@dataclass
class EndpointUpdaterSpec(UpdaterSpec[EndpointRow]):
    """UpdaterSpec for endpoint updates."""

    resource_slots: OptionalState[ResourceSlot] = field(
        default_factory=OptionalState[ResourceSlot].nop
    )
    resource_opts: TriState[dict[str, Any]] = field(default_factory=TriState[dict[str, Any]].nop)
    cluster_mode: OptionalState[ClusterMode] = field(default_factory=OptionalState[ClusterMode].nop)
    cluster_size: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    replicas: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    desired_session_count: OptionalState[int] = field(default_factory=OptionalState.nop)
    image: TriState[ImageRef] = field(default_factory=TriState.nop)
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    resource_group: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    model_definition_path: TriState[str] = field(default_factory=TriState[str].nop)
    open_to_public: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)
    extra_mounts: OptionalState[list[ExtraMount]] = field(
        default_factory=OptionalState[list[ExtraMount]].nop
    )
    environ: TriState[dict[str, str]] = field(default_factory=TriState[dict[str, str]].nop)
    runtime_variant: OptionalState[RuntimeVariant] = field(
        default_factory=OptionalState[RuntimeVariant].nop
    )

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.resource_slots.update_dict(to_update, "resource_slots")
        self.resource_opts.update_dict(to_update, "resource_opts")
        self.cluster_mode.update_dict(to_update, "cluster_mode")
        self.cluster_size.update_dict(to_update, "cluster_size")
        self.model_definition_path.update_dict(to_update, "model_definition_path")
        self.runtime_variant.update_dict(to_update, "runtime_variant")
        self.resource_group.update_dict(to_update, "resource_group")
        self.desired_session_count.update_dict(to_update, "desired_session_count")
        self.replicas.update_dict(to_update, "replicas")
        self.environ.update_dict(to_update, "environ")
        return to_update

    def replica_count_modified(self) -> bool:
        """Check if replicas field was modified."""
        return self.replicas.optional_value() is not None


@dataclass
class EndpointAutoScalingRuleUpdaterSpec(UpdaterSpec[EndpointAutoScalingRuleRow]):
    """UpdaterSpec for endpoint auto scaling rule updates."""

    metric_source: OptionalState[AutoScalingMetricSource] = field(default_factory=OptionalState.nop)
    metric_name: OptionalState[str] = field(default_factory=OptionalState.nop)
    threshold: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    comparator: OptionalState[AutoScalingMetricComparator] = field(
        default_factory=OptionalState.nop
    )
    step_size: OptionalState[int] = field(default_factory=OptionalState.nop)
    cooldown_seconds: OptionalState[int] = field(default_factory=OptionalState.nop)
    min_replicas: TriState[int] = field(default_factory=TriState.nop)
    max_replicas: TriState[int] = field(default_factory=TriState.nop)

    @property
    @override
    def row_class(self) -> type[EndpointAutoScalingRuleRow]:
        return EndpointAutoScalingRuleRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.metric_source.update_dict(to_update, "metric_source")
        self.metric_name.update_dict(to_update, "metric_name")
        self.threshold.update_dict(to_update, "threshold")
        self.comparator.update_dict(to_update, "comparator")
        self.step_size.update_dict(to_update, "step_size")
        self.cooldown_seconds.update_dict(to_update, "cooldown_seconds")
        self.min_replicas.update_dict(to_update, "min_replicas")
        self.max_replicas.update_dict(to_update, "max_replicas")
        return to_update



"""Factory functions for creating scheduler components and coordinator handlers."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.manager.clients.agent import AgentClientPool
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.plugin.network import NetworkPluginContext
from ai.backend.manager.repositories.deployment.repository import DeploymentRepository
from ai.backend.manager.repositories.fair_share import FairShareRepository
from ai.backend.manager.repositories.resource_usage_history import (
    ResourceUsageHistoryRepository,
)
from ai.backend.manager.repositories.scheduler import SchedulerRepository
from ai.backend.manager.scheduler.types import ScheduleType

if TYPE_CHECKING:
    from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import KernelMatchType, SessionStatus
from ai.backend.manager.sokovan.data import PromotionSpec
from ai.backend.manager.sokovan.scheduler.fair_share import (
    FairShareAggregator,
    FairShareFactorCalculator,
)
from ai.backend.manager.sokovan.scheduler.handlers import (
    CheckPreconditionLifecycleHandler,
    DeprioritizeSessionsLifecycleHandler,
    ScheduleSessionsLifecycleHandler,
    SessionLifecycleHandler,
    StartSessionsLifecycleHandler,
    SweepSessionsLifecycleHandler,
    TerminateSessionsLifecycleHandler,
)
from ai.backend.manager.sokovan.scheduler.handlers.kernel import (
    KernelLifecycleHandler,
    SweepStaleKernelsKernelHandler,
)
from ai.backend.manager.sokovan.scheduler.handlers.observer import (
    FairShareObserver,
    KernelObserver,
)
from ai.backend.manager.sokovan.scheduler.launcher.launcher import (
    SessionLauncher,
    SessionLauncherArgs,
)
from ai.backend.manager.sokovan.scheduler.provisioner.allocators.repository_allocator import (
    RepositoryAllocator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.provisioner import (
    SessionProvisioner,
    SessionProvisionerArgs,
)
from ai.backend.manager.sokovan.scheduler.provisioner.selectors.concentrated import (
    ConcentratedAgentSelector,
)
from ai.backend.manager.sokovan.scheduler.provisioner.selectors.selector import AgentSelector
from ai.backend.manager.sokovan.scheduler.provisioner.sequencers.fifo import FIFOSequencer
from ai.backend.manager.sokovan.scheduler.provisioner.validators.concurrency import (
    ConcurrencyValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.dependencies import (
    DependenciesValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.domain_resource_limit import (
    DomainResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.group_resource_limit import (
    GroupResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.keypair_resource_limit import (
    KeypairResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.pending_session_count_limit import (
    PendingSessionCountLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.pending_session_resource_limit import (
    PendingSessionResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.reserved_batch import (
    ReservedBatchSessionValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.user_resource_limit import (
    UserResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.provisioner.validators.validator import (
    SchedulingValidator,
)
from ai.backend.manager.sokovan.scheduler.scheduler import (
    SchedulerComponents,
    create_scheduler_components,
)
from ai.backend.manager.sokovan.scheduler.terminator.terminator import (
    SessionTerminator,
    SessionTerminatorArgs,
)


def create_default_scheduler_components(
    repository: SchedulerRepository,
    deployment_repository: DeploymentRepository,
    fair_share_repository: FairShareRepository,
    config_provider: ManagerConfigProvider,
    agent_client_pool: AgentClientPool,
    network_plugin_ctx: NetworkPluginContext,
    event_producer: EventProducer,
    valkey_schedule: ValkeyScheduleClient,
) -> SchedulerComponents:
    """
    Create scheduler components with default configuration.

    Args:
        repository: The repository for accessing system data
        deployment_repository: The deployment repository
        fair_share_repository: The fair share repository for sequencing
        config_provider: The manager configuration provider
        agent_client_pool: Pool for managing agent clients
        network_plugin_ctx: Network plugin context for network management
        event_producer: Event producer for publishing events
        valkey_schedule: Valkey client for scheduling operations

    Returns:
        A configured SchedulerComponents instance
    """
    # Create provisioner components
    sequencer = FIFOSequencer()
    validator = SchedulingValidator([
        ConcurrencyValidator(),
        DependenciesValidator(),
        DomainResourceLimitValidator(),
        GroupResourceLimitValidator(),
        KeypairResourceLimitValidator(),
        PendingSessionCountLimitValidator(),
        PendingSessionResourceLimitValidator(),
        ReservedBatchSessionValidator(),
        UserResourceLimitValidator(),
    ])
    resource_priority = config_provider.config.manager.agent_selection_resource_priority
    agent_selector = AgentSelector(ConcentratedAgentSelector(resource_priority))
    allocator = RepositoryAllocator(repository)

    # Create provisioner
    provisioner = SessionProvisioner(
        SessionProvisionerArgs(
            validator=validator,
            default_sequencer=sequencer,
            default_agent_selector=agent_selector,
            allocator=allocator,
            repository=repository,
            fair_share_repository=fair_share_repository,
            config_provider=config_provider,
            valkey_schedule=valkey_schedule,
        )
    )

    # Create launcher
    launcher = SessionLauncher(
        SessionLauncherArgs(
            repository=repository,
            agent_client_pool=agent_client_pool,
            network_plugin_ctx=network_plugin_ctx,
            config_provider=config_provider,
            valkey_schedule=valkey_schedule,
        )
    )

    # Create terminator
    terminator = SessionTerminator(
        SessionTerminatorArgs(
            repository=repository,
            agent_client_pool=agent_client_pool,
            valkey_schedule=valkey_schedule,
        )
    )

    return create_scheduler_components(
        provisioner=provisioner,
        launcher=launcher,
        terminator=terminator,
        repository=repository,
        deployment_repository=deployment_repository,
        config_provider=config_provider,
        agent_client_pool=agent_client_pool,
        event_producer=event_producer,
    )


# =============================================================================
# Coordinator Handlers
# =============================================================================


@dataclass
class CoordinatorHandlers:
    """Container for all handlers and specs injected into Coordinator.

    This dataclass decouples the Coordinator from handler creation logic,
    allowing handlers to be created externally and injected.
    """

    lifecycle_handlers: Mapping[ScheduleType, SessionLifecycleHandler]
    promotion_specs: Mapping[ScheduleType, PromotionSpec]
    kernel_handlers: Mapping[ScheduleType, KernelLifecycleHandler]
    kernel_observers: Mapping[ScheduleType, KernelObserver]


@dataclass
class CoordinatorHandlersArgs:
    """Arguments for creating CoordinatorHandlers."""

    provisioner: SessionProvisioner
    launcher: SessionLauncher
    terminator: SessionTerminator
    repository: SchedulerRepository
    valkey_schedule: ValkeyScheduleClient
    scheduling_controller: SchedulingController
    fair_share_aggregator: FairShareAggregator
    fair_share_calculator: FairShareFactorCalculator
    resource_usage_repository: ResourceUsageHistoryRepository
    fair_share_repository: FairShareRepository


def create_coordinator_handlers(args: CoordinatorHandlersArgs) -> CoordinatorHandlers:
    """Create all handlers and specs for the Coordinator.

    This factory function centralizes handler creation, decoupling the
    Coordinator from the details of handler instantiation.
    """
    lifecycle_handlers = _create_lifecycle_handlers(args)
    promotion_specs = _create_promotion_specs()
    kernel_handlers = _create_kernel_handlers(args)
    kernel_observers = _create_kernel_observers(args)

    return CoordinatorHandlers(
        lifecycle_handlers=lifecycle_handlers,
        promotion_specs=promotion_specs,
        kernel_handlers=kernel_handlers,
        kernel_observers=kernel_observers,
    )


def _create_lifecycle_handlers(
    args: CoordinatorHandlersArgs,
) -> Mapping[ScheduleType, SessionLifecycleHandler]:
    """Create lifecycle handlers mapping."""
    return {
        ScheduleType.SCHEDULE: ScheduleSessionsLifecycleHandler(
            args.provisioner,
            args.repository,
        ),
        ScheduleType.DEPRIORITIZE: DeprioritizeSessionsLifecycleHandler(
            args.repository,
        ),
        ScheduleType.CHECK_PRECONDITION: CheckPreconditionLifecycleHandler(
            args.launcher,
            args.repository,
        ),
        ScheduleType.START: StartSessionsLifecycleHandler(
            args.launcher,
            args.repository,
        ),
        ScheduleType.TERMINATE: TerminateSessionsLifecycleHandler(
            args.terminator,
            args.repository,
        ),
        ScheduleType.SWEEP: SweepSessionsLifecycleHandler(
            args.repository,
        ),
    }


def _create_promotion_specs() -> Mapping[ScheduleType, PromotionSpec]:
    """Create promotion specs mapping.

    Kernel matching semantics:
    - NOT_ANY: None of the kernels should be in these statuses (used for "no longer in X" checks)
    - ANY: At least one kernel should be in these statuses (used for early detection)
    - ALL: All kernels must be in these statuses (used for completion checks)
    """
    return {
        # Promote to PREPARED when no kernel is in pre-prepared states
        # (handles terminal statuses gracefully)
        ScheduleType.CHECK_PULLING_PROGRESS: PromotionSpec(
            name="promote-to-prepared",
            target_statuses=[SessionStatus.PREPARING, SessionStatus.PULLING],
            target_kernel_statuses=list(KernelStatus.pre_prepared_statuses()),
            kernel_match_type=KernelMatchType.NOT_ANY,
            success_status=SessionStatus.PREPARED,
            reason="triggered-by-scheduler",
        ),
        # Promote to RUNNING when no kernel is in pre-running states
        # (allows partial failure - some kernels may be TERMINATED)
        ScheduleType.CHECK_CREATING_PROGRESS: PromotionSpec(
            name="promote-to-running",
            target_statuses=[SessionStatus.CREATING],
            target_kernel_statuses=list(KernelStatus.pre_running_statuses()),
            kernel_match_type=KernelMatchType.NOT_ANY,
            success_status=SessionStatus.RUNNING,
            reason="triggered-by-scheduler",
        ),
        # Promote to TERMINATED when all kernels are TERMINATED
        ScheduleType.CHECK_TERMINATING_PROGRESS: PromotionSpec(
            name="promote-to-terminated",
            target_statuses=[SessionStatus.TERMINATING],
            target_kernel_statuses=[KernelStatus.TERMINATED],
            kernel_match_type=KernelMatchType.ALL,
            success_status=SessionStatus.TERMINATED,
            reason="triggered-by-scheduler",
        ),
        # Detect abnormal termination when ANY kernel is TERMINATED or CANCELLED
        # Covers all active session states where kernels can be terminated
        ScheduleType.DETECT_KERNEL_TERMINATION: PromotionSpec(
            name="detect-termination",
            target_statuses=[
                SessionStatus.PENDING,
                SessionStatus.SCHEDULED,
                SessionStatus.PREPARING,
                SessionStatus.PULLING,
                SessionStatus.PREPARED,
                SessionStatus.CREATING,
                SessionStatus.RUNNING,
                SessionStatus.DEPRIORITIZING,
            ],
            target_kernel_statuses=[KernelStatus.TERMINATED, KernelStatus.CANCELLED],
            kernel_match_type=KernelMatchType.ANY,
            success_status=SessionStatus.TERMINATING,
            reason="ABNORMAL_TERMINATION",
        ),
    }


def _create_kernel_handlers(
    args: CoordinatorHandlersArgs,
) -> Mapping[ScheduleType, KernelLifecycleHandler]:
    """Create kernel handlers mapping."""
    return {
        ScheduleType.SWEEP_STALE_KERNELS: SweepStaleKernelsKernelHandler(
            args.terminator,
        ),
    }


def _create_kernel_observers(
    args: CoordinatorHandlersArgs,
) -> Mapping[ScheduleType, KernelObserver]:
    """Create kernel observers mapping.

    Observers are read-only operations that collect data without changing
    kernel state. Used for fair share usage tracking, metrics, etc.
    """
    return {
        ScheduleType.OBSERVE_FAIR_SHARE: FairShareObserver(
            aggregator=args.fair_share_aggregator,
            calculator=args.fair_share_calculator,
            resource_usage_repository=args.resource_usage_repository,
            fair_share_repository=args.fair_share_repository,
            scheduler_repository=args.repository,
        ),
    }



"""kernel_requested_slots_to_not_nullable

Revision ID: 85615e005fa3
Revises: a083c6c962e5
Create Date: 2023-09-22 12:37:31.725324

"""

from collections.abc import Sequence
from typing import Any

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.engine import Row

from ai.backend.manager.models.base import GUID, mapper_registry

# revision identifiers, used by Alembic.
revision = "85615e005fa3"
down_revision = "a083c6c962e5"
branch_labels = None
depends_on = None


BATCH_SIZE = 100


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    connection = op.get_bind()
    kernels = sa.Table(
        "kernels",
        mapper_registry.metadata,
        sa.Column(
            "id",
            GUID,
            primary_key=True,
        ),
        sa.Column("requested_slots", pgsql.JSONB()),
        extend_existing=True,
    )

    while True:
        select_stmt = (
            sa.select(kernels.c.id).where(kernels.c.requested_slots.is_(None)).limit(BATCH_SIZE)
        )
        fetch_result: Sequence[Row[Any]] = connection.execute(select_stmt).fetchall()
        kernel_ids_to_update = [row.id for row in fetch_result]

        stmt = (
            sa.update(kernels)
            .where(kernels.c.id.in_(kernel_ids_to_update))
            .values(requested_slots={})
        )
        connection.execute(stmt)
        if len(kernel_ids_to_update) < BATCH_SIZE:
            break
    op.alter_column("kernels", "requested_slots", nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("kernels", "requested_slots", nullable=True)
    # ### end Alembic commands ###



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.artifact_registries.types import ArtifactRegistryModifierMeta
from ai.backend.manager.data.reservoir_registry.types import ReservoirRegistryData
from ai.backend.manager.models.reservoir_registry import ReservoirRegistryRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.artifact_registry.actions.base import ArtifactRegistryAction


@dataclass
class UpdateReservoirRegistryAction(ArtifactRegistryAction):
    updater: Updater[ReservoirRegistryRow]
    meta: ArtifactRegistryModifierMeta

    @override
    def entity_id(self) -> str | None:
        return str(self.updater.pk_value)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpdateReservoirRegistryActionResult(BaseActionResult):
    result: ReservoirRegistryData

    @override
    def entity_id(self) -> str | None:
        return str(self.result.id)



"""Database source for scaling group repository operations."""

from __future__ import annotations

import uuid
from decimal import Decimal
from typing import TYPE_CHECKING, cast

import sqlalchemy as sa

from ai.backend.common.exception import ScalingGroupConflict
from ai.backend.common.types import ResourceSlot
from ai.backend.manager.data.agent.types import AgentStatus
from ai.backend.manager.data.scaling_group.types import (
    ResourceInfo,
    ScalingGroupData,
    ScalingGroupListResult,
)
from ai.backend.manager.errors.resource import ScalingGroupNotFound
from ai.backend.manager.models.agent import AgentRow
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.models.resource_slot import AgentResourceRow
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.models.scaling_group import (
    ScalingGroupForDomainRow,
    ScalingGroupForKeypairsRow,
    ScalingGroupForProjectRow,
    ScalingGroupRow,
)
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.repositories.base import BatchQuerier, execute_batch_querier
from ai.backend.manager.repositories.base.creator import (
    BulkCreator,
    Creator,
    execute_bulk_creator,
    execute_creator,
)
from ai.backend.manager.repositories.base.purger import (
    BatchPurger,
    Purger,
    execute_batch_purger,
    execute_purger,
)
from ai.backend.manager.repositories.base.updater import Updater, execute_updater
from ai.backend.manager.repositories.scaling_group.creators import ScalingGroupCreatorSpec

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


__all__ = (
    "ScalingGroupDBSource",
    "ScalingGroupListResult",
)


class ScalingGroupDBSource:
    """
    Database source for scaling group operations.
    Handles all database operations for scaling groups.
    """

    _db: ExtendedAsyncSAEngine

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
    ) -> None:
        self._db = db

    async def create_scaling_group(
        self,
        creator: Creator[ScalingGroupRow],
    ) -> ScalingGroupData:
        """Creates a new scaling group.

        Raises ScalingGroupConflict if a scaling group with the same name already exists.
        """
        async with self._db.begin_session() as session:
            try:
                result = await execute_creator(session, creator)
            except sa.exc.IntegrityError as e:
                spec = cast(ScalingGroupCreatorSpec, creator.spec)
                raise ScalingGroupConflict(f"Duplicate scaling group name: {spec.name}") from e
            return result.row.to_dataclass()

    async def search_scaling_groups(
        self,
        querier: BatchQuerier,
    ) -> ScalingGroupListResult:
        """Searches scaling groups with total count."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ScalingGroupRow)

            result = await execute_batch_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.ScalingGroupRow.to_dataclass() for row in result.rows]

            return ScalingGroupListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def get_scaling_group_by_name(
        self,
        name: str,
    ) -> ScalingGroupData:
        """Get a single scaling group by name (primary key).

        Args:
            name: The name of the scaling group (primary key).

        Returns:
            ScalingGroupData for the requested scaling group.

        Raises:
            ScalingGroupNotFound: If the scaling group does not exist.
        """
        async with self._db.begin_readonly_session_read_committed() as db_sess:
            row = await db_sess.get(ScalingGroupRow, name)
            if row is None:
                raise ScalingGroupNotFound(name)
            return row.to_dataclass()

    async def purge_scaling_group(
        self,
        purger: Purger[ScalingGroupRow],
    ) -> ScalingGroupData:
        """Purges a scaling group and all related sessions and routes using a purger.

        Cascade delete order:
        1. RoutingRow (session FK with RESTRICT)
        2. EndpointRow (resource_group FK with RESTRICT, has CASCADE to routing)
        3. SessionRow (scaling_group FK)
        4. ScalingGroupRow

        Raises ScalingGroupNotFound if scaling group doesn't exist.
        """
        async with self._db.begin_session() as session:
            scaling_group_name = purger.pk_value

            # Step 1: Find all sessions belonging to this scaling group
            session_ids_query = sa.select(SessionRow.id).where(
                SessionRow.scaling_group_name == scaling_group_name
            )
            session_ids_result = await session.execute(session_ids_query)
            session_ids = session_ids_result.scalars().all()

            # Step 2: Delete all routings associated with these sessions
            if session_ids:
                delete_routings_stmt = sa.delete(RoutingRow).where(
                    RoutingRow.session.in_(session_ids)
                )
                await session.execute(delete_routings_stmt)

            # Step 3: Delete all endpoints belonging to this scaling group
            delete_endpoints_stmt = sa.delete(EndpointRow).where(
                EndpointRow.resource_group == scaling_group_name
            )
            await session.execute(delete_endpoints_stmt)

            # Step 4: Delete all sessions belonging to this scaling group
            delete_sessions_stmt = sa.delete(SessionRow).where(
                SessionRow.scaling_group_name == scaling_group_name
            )
            await session.execute(delete_sessions_stmt)

            # Step 5: Delete the scaling group itself using purger
            result = await execute_purger(session, purger)

            if result is None:
                raise ScalingGroupNotFound(f"Scaling group not found (name:{purger.pk_value})")

            return result.row.to_dataclass()

    async def update_scaling_group(
        self,
        updater: Updater[ScalingGroupRow],
    ) -> ScalingGroupData:
        """Updates an existing scaling group.

        Raises ScalingGroupNotFound if the scaling group does not exist.
        """
        async with self._db.begin_session() as session:
            result = await execute_updater(session, updater)
            if result is None:
                raise ScalingGroupNotFound(f"Scaling group not found (name:{updater.pk_value})")
            return result.row.to_dataclass()

    async def associate_scaling_group_with_domains(
        self,
        bulk_creator: BulkCreator[ScalingGroupForDomainRow],
    ) -> None:
        """Associates a scaling group with multiple domains."""
        async with self._db.begin_session() as session:
            await execute_bulk_creator(session, bulk_creator)

    async def disassociate_scaling_group_with_domains(
        self,
        purger: BatchPurger[ScalingGroupForDomainRow],
    ) -> None:
        """Disassociates a scaling group from multiple domains."""
        async with self._db.begin_session() as session:
            await execute_batch_purger(session, purger)

    async def check_scaling_group_domain_association_exists(
        self,
        scaling_group: str,
        domain: str,
    ) -> bool:
        """Checks if a scaling group is associated with a domain."""
        async with self._db.begin_readonly_session_read_committed() as session:
            query = (
                sa.select(sa.func.count())
                .select_from(ScalingGroupForDomainRow)
                .where(
                    sa.and_(
                        ScalingGroupForDomainRow.scaling_group == scaling_group,
                        ScalingGroupForDomainRow.domain == domain,
                    )
                )
            )
            result = await session.scalar(query)
            return (result or 0) > 0

    async def associate_scaling_group_with_keypairs(
        self,
        bulk_creator: BulkCreator[ScalingGroupForKeypairsRow],
    ) -> None:
        """Associates a scaling group with multiple keypairs."""
        async with self._db.begin_session() as session:
            await execute_bulk_creator(session, bulk_creator)

    async def disassociate_scaling_group_with_keypairs(
        self,
        purger: BatchPurger[ScalingGroupForKeypairsRow],
    ) -> None:
        """Disassociates a scaling group from multiple keypairs."""
        async with self._db.begin_session() as session:
            await execute_batch_purger(session, purger)

    async def check_scaling_group_keypair_association_exists(
        self,
        scaling_group_name: str,
        access_key: str,
    ) -> bool:
        """Checks if a scaling group is associated with a keypair."""
        async with self._db.begin_readonly_session_read_committed() as session:
            query = sa.select(
                sa.exists().where(
                    sa.and_(
                        ScalingGroupForKeypairsRow.scaling_group == scaling_group_name,
                        ScalingGroupForKeypairsRow.access_key == access_key,
                    )
                )
            )
            result = await session.execute(query)
            return result.scalar() or False

    async def associate_scaling_group_with_user_groups(
        self,
        bulk_creator: BulkCreator[ScalingGroupForProjectRow],
    ) -> None:
        """Associates a scaling group with multiple user groups (projects)."""
        async with self._db.begin_session() as session:
            await execute_bulk_creator(session, bulk_creator)

    async def disassociate_scaling_group_with_user_groups(
        self,
        purger: BatchPurger[ScalingGroupForProjectRow],
    ) -> None:
        """Disassociates a single scaling group from a user group (project)."""
        async with self._db.begin_session() as session:
            await execute_batch_purger(session, purger)

    async def check_scaling_group_user_group_association_exists(
        self,
        scaling_group: str,
        user_group: uuid.UUID,
    ) -> bool:
        """Checks if a scaling group is associated with a user group (project)."""
        async with self._db.begin_readonly_session_read_committed() as session:
            query = (
                sa.select(sa.func.count())
                .select_from(ScalingGroupForProjectRow)
                .where(
                    sa.and_(
                        ScalingGroupForProjectRow.scaling_group == scaling_group,
                        ScalingGroupForProjectRow.group == user_group,
                    )
                )
            )
            result = await session.scalar(query)
            return (result or 0) > 0

    async def get_resource_info(
        self,
        scaling_group: str,
    ) -> ResourceInfo:
        """Get aggregated resource information for a scaling group.

        Uses normalized agent_resources table with SQL-level aggregation.

        Args:
            scaling_group: The name of the scaling group.

        Returns:
            ResourceInfo containing capacity, used, and free resource metrics.

        Raises:
            ScalingGroupNotFound: If the scaling group does not exist.
        """
        ar = AgentResourceRow.__table__
        ag = AgentRow.__table__

        async with self._db.begin_readonly_session() as db_sess:
            # Validate scaling group exists
            sg_exists = await db_sess.scalar(
                sa.select(sa.exists().where(ScalingGroupRow.name == scaling_group))
            )
            if not sg_exists:
                raise ScalingGroupNotFound(scaling_group)

            # Capacity: ALIVE + schedulable agents
            capacity_stmt = (
                sa.select(ar.c.slot_name, sa.func.sum(ar.c.capacity).label("total"))
                .select_from(ar.join(ag, ar.c.agent_id == ag.c.id))
                .where(
                    ag.c.scaling_group == scaling_group,
                    ag.c.status == AgentStatus.ALIVE,
                    ag.c.schedulable == sa.true(),
                )
                .group_by(ar.c.slot_name)
            )
            capacity_result = await db_sess.execute(capacity_stmt)
            capacity_dict: dict[str, Decimal] = {
                row.slot_name: row.total for row in capacity_result
            }

            # Used: ALIVE agents (regardless of schedulable)
            used_stmt = (
                sa.select(ar.c.slot_name, sa.func.sum(ar.c.used).label("total"))
                .select_from(ar.join(ag, ar.c.agent_id == ag.c.id))
                .where(
                    ag.c.scaling_group == scaling_group,
                    ag.c.status == AgentStatus.ALIVE,
                )
                .group_by(ar.c.slot_name)
            )
            used_result = await db_sess.execute(used_stmt)
            used_dict: dict[str, Decimal] = {row.slot_name: row.total for row in used_result}

        capacity = ResourceSlot({k: v for k, v in capacity_dict.items()})
        used = ResourceSlot({k: v for k, v in used_dict.items()})
        free = capacity - used

        return ResourceInfo(
            capacity=capacity,
            used=used,
            free=free,
        )



"""add user_resource_policies.max_session_count_per_model_session

Revision ID: 75ea2b136830
Revises: 589c764a18f1
Create Date: 2024-03-04 22:21:49.888739

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "75ea2b136830"
down_revision = "589c764a18f1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "user_resource_policies",
        sa.Column("max_session_count_per_model_session", sa.Integer(), nullable=True, default=10),
    )
    op.execute("UPDATE user_resource_policies SET max_session_count_per_model_session = 10")
    op.alter_column("user_resource_policies", "max_session_count_per_model_session", nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("user_resource_policies", "max_session_count_per_model_session")
    # ### end Alembic commands ###



from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.services.container_registry.actions.base import ContainerRegistryAction


@dataclass
class ClearImagesAction(ContainerRegistryAction):
    registry: str
    project: str | None

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.CONTAINER_REGISTRY_IMAGE

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class ClearImagesActionResult(BaseActionResult):
    registry: ContainerRegistryData

    @override
    def entity_id(self) -> str | None:
        return str(self.registry.id)


# @dataclass
# class ClearImagesBatchAction(ImageAction):
#     registry: str

#     @override
#     def entity_id(self) -> Optional[str]:
#         return None

#     @override
#     def operation_type(self):
#         return "clear_multi"


# @dataclass
# class ClearImagesBatchActionResult(BaseActionResult):
#     registry_row: ContainerRegistryRow

#     @override
#     def entity_id(self) -> Optional[str]:
#         return None



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.scaling_group.types import ScalingGroupData
from ai.backend.manager.models.scaling_group import ScalingGroupRow
from ai.backend.manager.repositories.base.purger import Purger
from ai.backend.manager.services.scaling_group.actions.base import ScalingGroupAction


@dataclass
class PurgeScalingGroupAction(ScalingGroupAction):
    """Action to purge a scaling group by name, including all related sessions and routes."""

    purger: Purger[ScalingGroupRow]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.PURGE

    @override
    def entity_id(self) -> str | None:
        return str(self.purger.pk_value)


@dataclass
class PurgeScalingGroupActionResult(BaseActionResult):
    """Result of purging a scaling group."""

    data: ScalingGroupData

    @override
    def entity_id(self) -> str | None:
        return self.data.name



"""merge f83d and 1f55

Revision ID: 360af8f33d4e
Revises: f83d630c0bc9, 1f55a65cfc4f
Create Date: 2022-10-02 12:03:01.116361

"""

# revision identifiers, used by Alembic.
revision = "360af8f33d4e"
down_revision = ("f83d630c0bc9", "1f55a65cfc4f")
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



"""Query conditions and orders for auto-scaling rules."""

from __future__ import annotations

import uuid
from collections.abc import Collection
from datetime import datetime

import sqlalchemy as sa

from ai.backend.manager.models.endpoint import EndpointAutoScalingRuleRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class AutoScalingRuleConditions:
    """Query conditions for auto-scaling rules."""

    @staticmethod
    def by_ids(rule_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.id.in_(rule_ids)

        return inner

    @staticmethod
    def by_endpoint_id(endpoint_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.endpoint == endpoint_id

        return inner

    @staticmethod
    def by_deployment_id(deployment_id: uuid.UUID) -> QueryCondition:
        """Alias for by_endpoint_id for consistency with deployment naming."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.endpoint == deployment_id

        return inner

    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.created_at == dt

        return inner

    @staticmethod
    def by_last_triggered_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.last_triggered_at < dt

        return inner

    @staticmethod
    def by_last_triggered_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.last_triggered_at > dt

        return inner

    @staticmethod
    def by_last_triggered_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointAutoScalingRuleRow.last_triggered_at == dt

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(EndpointAutoScalingRuleRow.created_at)
                .where(EndpointAutoScalingRuleRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return EndpointAutoScalingRuleRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(EndpointAutoScalingRuleRow.created_at)
                .where(EndpointAutoScalingRuleRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return EndpointAutoScalingRuleRow.created_at > subquery

        return inner


class AutoScalingRuleOrders:
    """Query orders for auto-scaling rules."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return EndpointAutoScalingRuleRow.created_at.asc()
        return EndpointAutoScalingRuleRow.created_at.desc()



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

import strawberry

from ai.backend.manager.api.gql.data_loader.data_loaders import DataLoaders
from ai.backend.manager.api.gql.data_loader.registry import DataLoaderRegistry
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.services.processors import Processors

if TYPE_CHECKING:
    from ai.backend.common.events.fetcher import EventFetcher
    from ai.backend.common.events.hub.hub import EventHub
    from ai.backend.manager.api.gql.adapter import BaseGQLAdapter


class GQLFilter(ABC):
    """Abstract base class for GraphQL filter types.

    All GraphQL filter input types should inherit from this ABC
    to ensure they implement the build_conditions method.
    """

    @abstractmethod
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns:
            A list of QueryCondition callables that can be applied to SQLAlchemy queries.
        """
        raise NotImplementedError


class GQLOrderBy(ABC):
    """Abstract base class for GraphQL order by types.

    All GraphQL order by input types should inherit from this ABC
    to ensure they implement the to_query_order method.
    """

    @abstractmethod
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder.

        Returns:
            A QueryOrder (SQLAlchemy UnaryExpression) for ordering query results.
        """
        raise NotImplementedError


@dataclass
class StrawberryGQLContext:
    processors: Processors
    config_provider: ManagerConfigProvider
    event_hub: EventHub
    event_fetcher: EventFetcher
    dataloader_registry: DataLoaderRegistry  # TODO: Remove this.
    gql_adapter: BaseGQLAdapter
    data_loaders: DataLoaders


# Scope input types for BEP-1041 Resource Group scoped APIs


@strawberry.input(description="Resource group scope for domain-level operations")
class ResourceGroupDomainScope:
    """Scope for domain-level APIs within a resource group context."""

    resource_group_name: str = strawberry.field(
        description="Resource group name to scope the operation"
    )


@strawberry.input(description="Resource group + domain scope for project-level operations")
class ResourceGroupProjectScope:
    """Scope for project-level APIs within a resource group and domain context."""

    resource_group_name: str = strawberry.field(
        description="Resource group name to scope the operation"
    )
    domain_name: str = strawberry.field(description="Domain name to scope the operation")


@strawberry.input(description="Resource group + domain + project scope for user-level operations")
class ResourceGroupUserScope:
    """Scope for user-level APIs within a resource group, domain, and project context."""

    resource_group_name: str = strawberry.field(
        description="Resource group name to scope the operation"
    )
    domain_name: str = strawberry.field(description="Domain name to scope the operation")
    project_id: str = strawberry.field(description="Project ID to scope the operation")


# Scope input types for Usage Bucket scoped APIs


@strawberry.input(description="Domain scope for usage bucket queries")
class DomainUsageBucketScope:
    """Scope for domain-level usage bucket APIs."""

    resource_group_name: str = strawberry.field(description="Resource group name")
    domain_name: str = strawberry.field(description="Domain name to retrieve usage buckets for")


@strawberry.input(description="Project scope for usage bucket queries")
class ProjectUsageBucketScope:
    """Scope for project-level usage bucket APIs."""

    resource_group_name: str = strawberry.field(description="Resource group name")
    domain_name: str = strawberry.field(description="Domain name")
    project_id: str = strawberry.field(description="Project ID (will be converted to UUID)")


@strawberry.input(description="User scope for usage bucket queries")
class UserUsageBucketScope:
    """Scope for user-level usage bucket APIs."""

    resource_group_name: str = strawberry.field(description="Resource group name")
    domain_name: str = strawberry.field(description="Domain name")
    project_id: str = strawberry.field(description="Project ID (will be converted to UUID)")
    user_uuid: str = strawberry.field(description="User UUID (will be converted to UUID)")



"""Scheduling context data types."""

from dataclasses import dataclass

from ai.backend.manager.sokovan.data import (
    ScalingGroupInfo,
    SchedulingConfig,
    SessionWorkload,
    SystemSnapshot,
)
from ai.backend.manager.sokovan.scheduler.provisioner.selectors.selector import AgentInfo


@dataclass
class SchedulingContextData:
    """Processed data ready for scheduling decisions."""

    scaling_group_info: ScalingGroupInfo
    pending_sessions: list[SessionWorkload]
    system_snapshot: SystemSnapshot
    scheduling_config: SchedulingConfig
    agents: list[AgentInfo]



import uuid

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel
from ai.backend.common.data.notification import NotificationRuleType


class UpdateNotificationChannelPathParam(BaseRequestModel):
    channel_id: uuid.UUID = Field(description="The notification channel ID to update")


class GetNotificationChannelPathParam(BaseRequestModel):
    channel_id: uuid.UUID = Field(description="The notification channel ID to retrieve")


class DeleteNotificationChannelPathParam(BaseRequestModel):
    channel_id: uuid.UUID = Field(description="The notification channel ID to delete")


class UpdateNotificationRulePathParam(BaseRequestModel):
    rule_id: uuid.UUID = Field(description="The notification rule ID to update")


class GetNotificationRulePathParam(BaseRequestModel):
    rule_id: uuid.UUID = Field(description="The notification rule ID to retrieve")


class DeleteNotificationRulePathParam(BaseRequestModel):
    rule_id: uuid.UUID = Field(description="The notification rule ID to delete")


class ValidateNotificationChannelPathParam(BaseRequestModel):
    channel_id: uuid.UUID = Field(description="The notification channel ID to validate")


class ValidateNotificationRulePathParam(BaseRequestModel):
    rule_id: uuid.UUID = Field(description="The notification rule ID to validate")


class RuleTypePathParam(BaseRequestModel):
    rule_type: NotificationRuleType = Field(
        description="The notification rule type to get schema for"
    )



"""User-level app configuration actions."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.app_config.types import AppConfigData
from ai.backend.manager.repositories.app_config.updaters import AppConfigUpdaterSpec

from .base import AppConfigAction


@dataclass
class GetUserConfigAction(AppConfigAction):
    """Action to get user-level app configuration."""

    user_id: str

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG_USER

    @override
    def entity_id(self) -> str | None:
        return self.user_id

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetUserConfigActionResult(BaseActionResult):
    """Result of get user config action."""

    result: AppConfigData | None

    @override
    def entity_id(self) -> str | None:
        return self.result.scope_id if self.result else None


@dataclass
class UpsertUserConfigAction(AppConfigAction):
    """Action to create or update user-level app configuration."""

    user_id: str
    updater_spec: AppConfigUpdaterSpec

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG_USER

    @override
    def entity_id(self) -> str | None:
        return self.user_id

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpsertUserConfigActionResult(BaseActionResult):
    """Result of upsert user config action."""

    result: AppConfigData

    @override
    def entity_id(self) -> str | None:
        return self.result.scope_id


@dataclass
class DeleteUserConfigAction(AppConfigAction):
    """Action to delete user-level app configuration."""

    user_id: str

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.APP_CONFIG_USER

    @override
    def entity_id(self) -> str | None:
        return self.user_id

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DeleteUserConfigActionResult(BaseActionResult):
    """Result of delete user config action."""

    deleted: bool
    user_id: str

    @override
    def entity_id(self) -> str | None:
        return self.user_id if self.deleted else None



import logging
from pathlib import Path

from ai.backend.common.dto.storage.request import (
    PresignedDownloadObjectReq,
    PresignedUploadObjectReq,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.clients.storage_proxy.session_manager import StorageSessionManager
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.artifact.types import ArtifactStatus
from ai.backend.manager.errors.artifact import ArtifactNotApproved, ArtifactReadonly
from ai.backend.manager.errors.common import ServerMisconfiguredError
from ai.backend.manager.errors.object_storage import ObjectStorageOperationNotSupported
from ai.backend.manager.repositories.artifact.repository import ArtifactRepository
from ai.backend.manager.repositories.object_storage.repository import ObjectStorageRepository
from ai.backend.manager.repositories.storage_namespace.repository import StorageNamespaceRepository
from ai.backend.manager.services.object_storage.actions.create import (
    CreateObjectStorageAction,
    CreateObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.delete import (
    DeleteObjectStorageAction,
    DeleteObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.get import (
    GetObjectStorageAction,
    GetObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.get_download_presigned_url import (
    GetDownloadPresignedURLAction,
    GetDownloadPresignedURLActionResult,
)
from ai.backend.manager.services.object_storage.actions.get_upload_presigned_url import (
    GetUploadPresignedURLAction,
    GetUploadPresignedURLActionResult,
)
from ai.backend.manager.services.object_storage.actions.list import (
    ListObjectStorageAction,
    ListObjectStorageActionResult,
)
from ai.backend.manager.services.object_storage.actions.search import (
    SearchObjectStoragesAction,
    SearchObjectStoragesActionResult,
)
from ai.backend.manager.services.object_storage.actions.update import (
    UpdateObjectStorageAction,
    UpdateObjectStorageActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ObjectStorageService:
    _artifact_repository: ArtifactRepository
    _object_storage_repository: ObjectStorageRepository
    _storage_namespace_repository: StorageNamespaceRepository
    _storage_manager: StorageSessionManager
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        artifact_repository: ArtifactRepository,
        object_storage_repository: ObjectStorageRepository,
        storage_namespace_repository: StorageNamespaceRepository,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._artifact_repository = artifact_repository
        self._object_storage_repository = object_storage_repository
        self._storage_namespace_repository = storage_namespace_repository
        self._storage_manager = storage_manager
        self._config_provider = config_provider

    async def create(self, action: CreateObjectStorageAction) -> CreateObjectStorageActionResult:
        """
        Create a new object storage.
        """
        log.info("Creating object storage with data: {}", action.creator)
        storage_data = await self._object_storage_repository.create(action.creator)
        return CreateObjectStorageActionResult(result=storage_data)

    async def update(self, action: UpdateObjectStorageAction) -> UpdateObjectStorageActionResult:
        """
        Update an existing object storage.
        """
        log.info("Updating object storage with id: {}", action.updater.pk_value)
        storage_data = await self._object_storage_repository.update(action.updater)
        return UpdateObjectStorageActionResult(result=storage_data)

    async def delete(self, action: DeleteObjectStorageAction) -> DeleteObjectStorageActionResult:
        """
        Delete an existing object storage.
        """
        log.info("Deleting object storage with id: {}", action.storage_id)
        storage_data = await self._object_storage_repository.delete(action.storage_id)
        return DeleteObjectStorageActionResult(deleted_storage_id=storage_data)

    async def get(self, action: GetObjectStorageAction) -> GetObjectStorageActionResult:
        """
        Get an existing object storage by ID.
        """
        log.info("Getting object storage with id: {}", action.storage_id)
        storage_data = await self._object_storage_repository.get_by_id(action.storage_id)
        return GetObjectStorageActionResult(result=storage_data)

    # TODO: Add filtering logic
    async def list(self, _action: ListObjectStorageAction) -> ListObjectStorageActionResult:
        """
        List all object storages.
        """
        log.info("Listing object storages")
        storage_data_list = await self._object_storage_repository.list_object_storages()
        return ListObjectStorageActionResult(data=storage_data_list)

    async def search(self, action: SearchObjectStoragesAction) -> SearchObjectStoragesActionResult:
        """Searches Object storages."""
        result = await self._object_storage_repository.search(
            querier=action.querier,
        )
        return SearchObjectStoragesActionResult(
            storages=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_presigned_download_url(
        self, action: GetDownloadPresignedURLAction
    ) -> GetDownloadPresignedURLActionResult:
        """
        Get a presigned download URL for an existing object storage.
        """
        log.info(
            "Getting presigned download URL for object storage, artifact_revision: {}",
            action.artifact_revision_id,
        )

        reservoir_config = self._config_provider.config.reservoir
        if reservoir_config is None:
            raise ServerMisconfiguredError("Reservoir configuration is missing")

        storage_name = reservoir_config.archive_storage

        # Only object storage is supported for presigned URLs
        if reservoir_config.config.storage_type != "object_storage":
            raise ObjectStorageOperationNotSupported(
                "Presigned URLs are only supported for object storage"
            )

        bucket_name = reservoir_config.config.bucket_name
        storage_data = await self._object_storage_repository.get_by_name(storage_name)
        storage_namespace = await self._storage_namespace_repository.get_by_storage_and_namespace(
            storage_data.id, bucket_name
        )
        revision_data = await self._artifact_repository.get_artifact_revision_by_id(
            action.artifact_revision_id
        )
        artifact_data = await self._artifact_repository.get_artifact_by_id(
            revision_data.artifact_id
        )

        if revision_data.status != ArtifactStatus.AVAILABLE:
            raise ArtifactNotApproved("Only available artifacts can be downloaded.")

        storage_proxy_client = self._storage_manager.get_manager_facing_client(storage_data.host)

        # Build S3 key
        object_path = Path(artifact_data.name) / revision_data.version / action.key

        result = await storage_proxy_client.get_s3_presigned_download_url(
            storage_data.name,
            storage_namespace.namespace,
            PresignedDownloadObjectReq(key=str(object_path), expiration=action.expiration),
        )

        return GetDownloadPresignedURLActionResult(
            storage_id=storage_data.id, presigned_url=result.url
        )

    async def get_presigned_upload_url(
        self, action: GetUploadPresignedURLAction
    ) -> GetUploadPresignedURLActionResult:
        """
        Get a presigned upload URL for an existing object storage.
        """
        log.info(
            "Getting presigned upload URL for object storage with artifact id: {}",
            action.artifact_revision_id,
        )

        reservoir_config = self._config_provider.config.reservoir
        if reservoir_config is None:
            raise ServerMisconfiguredError("Reservoir configuration is missing")

        storage_name = reservoir_config.archive_storage

        # Only object storage is supported for presigned URLs
        if reservoir_config.config.storage_type != "object_storage":
            raise ObjectStorageOperationNotSupported(
                "Presigned URLs are only supported for object storage"
            )

        bucket_name = reservoir_config.config.bucket_name
        storage_data = await self._object_storage_repository.get_by_name(storage_name)
        storage_namespace = await self._storage_namespace_repository.get_by_storage_and_namespace(
            storage_data.id, bucket_name
        )

        revision_data = await self._artifact_repository.get_artifact_revision_by_id(
            action.artifact_revision_id
        )
        artifact_data = await self._artifact_repository.get_artifact_by_id(
            revision_data.artifact_id
        )

        if artifact_data.readonly:
            raise ArtifactReadonly("Cannot upload to a readonly artifact.")

        storage_proxy_client = self._storage_manager.get_manager_facing_client(storage_data.host)

        # Build S3 key
        object_path = Path(artifact_data.name) / revision_data.version / action.key

        result = await storage_proxy_client.get_s3_presigned_upload_url(
            storage_data.name,
            storage_namespace.namespace,
            PresignedUploadObjectReq(
                key=str(object_path),
            ),
        )

        return GetUploadPresignedURLActionResult(
            storage_id=storage_data.id, presigned_url=result.url, fields=result.fields
        )



"""CreatorSpec for deployment (endpoint) creation."""

from __future__ import annotations

import uuid
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import Any, override

import yarl

from ai.backend.common.types import (
    ClusterMode,
    ResourceSlot,
    RuntimeVariant,
    VFolderMount,
)
from ai.backend.manager.models.endpoint import EndpointLifecycle, EndpointRow
from ai.backend.manager.repositories.base import CreatorSpec
from ai.backend.manager.repositories.base.updater import BatchUpdaterSpec

# ========== Field groups for DeploymentCreatorSpec ==========


@dataclass
class DeploymentMetadataFields:
    """Metadata fields for deployment creation.

    Corresponds to DeploymentMetadata in data layer.
    """

    name: str
    domain: str
    project_id: uuid.UUID
    resource_group: str
    created_user_id: uuid.UUID
    session_owner_id: uuid.UUID
    revision_history_limit: int = 10
    tag: str | None = None
    created_at: datetime | None = None


@dataclass
class DeploymentReplicaFields:
    """Replica fields for deployment creation.

    Corresponds to ReplicaSpec in data layer.
    """

    replica_count: int
    desired_replica_count: int | None = None


@dataclass
class DeploymentNetworkFields:
    """Network fields for deployment creation.

    Corresponds to DeploymentNetworkSpec in data layer.
    """

    open_to_public: bool = False
    url: str | None = None


@dataclass
class DeploymentResourceFields:
    """Resource fields for deployment model revision.

    Corresponds to ResourceSpec in data layer.
    """

    cluster_mode: ClusterMode
    cluster_size: int
    resource_slots: ResourceSlot
    resource_opts: Mapping[str, Any] | None = None


@dataclass
class DeploymentMountFields:
    """Mount fields for deployment model revision.

    Corresponds to VFolderMountsCreator in data layer.
    """

    model_vfolder_id: uuid.UUID | None
    model_mount_destination: str = "/models"
    model_definition_path: str | None = None
    extra_mounts: Sequence[VFolderMount] = ()


@dataclass
class DeploymentExecutionFields:
    """Execution fields for deployment model revision.

    Corresponds to ExecutionSpec in data layer.
    """

    runtime_variant: RuntimeVariant = RuntimeVariant.CUSTOM
    startup_command: str | None = None
    bootstrap_script: str | None = None
    environ: Mapping[str, str] | None = None
    callback_url: yarl.URL | None = None


@dataclass
class ModelRevisionFields:
    """Model revision fields for deployment creation.

    Corresponds to ModelRevisionCreator in data layer.
    Note: image_id is a resolved UUID, not ImageIdentifier.
    """

    image_id: uuid.UUID
    resource: DeploymentResourceFields
    mounts: DeploymentMountFields
    execution: DeploymentExecutionFields


@dataclass
class DeploymentCreatorSpec(CreatorSpec[EndpointRow]):
    """CreatorSpec for deployment creation with nested field groups.

    Corresponds to NewDeploymentCreator in data layer.
    All external references (like image) should be already resolved to UUIDs.
    """

    metadata: DeploymentMetadataFields
    replica: DeploymentReplicaFields
    network: DeploymentNetworkFields
    revision: ModelRevisionFields

    @override
    def build_row(self) -> EndpointRow:
        return EndpointRow(
            # Metadata fields
            name=self.metadata.name,
            domain=self.metadata.domain,
            project=self.metadata.project_id,
            resource_group=self.metadata.resource_group,
            created_user=self.metadata.created_user_id,
            session_owner=self.metadata.session_owner_id,
            revision_history_limit=self.metadata.revision_history_limit,
            tag=self.metadata.tag,
            # Replica fields
            replicas=self.replica.replica_count,
            desired_replicas=self.replica.desired_replica_count,
            # Network fields
            open_to_public=self.network.open_to_public,
            url=self.network.url,
            # Revision fields - image
            image=self.revision.image_id,
            # Revision fields - resource
            cluster_mode=self.revision.resource.cluster_mode,
            cluster_size=self.revision.resource.cluster_size,
            resource_slots=self.revision.resource.resource_slots,
            resource_opts=self.revision.resource.resource_opts or {},
            # Revision fields - mounts
            model=self.revision.mounts.model_vfolder_id,
            model_mount_destination=self.revision.mounts.model_mount_destination,
            model_definition_path=self.revision.mounts.model_definition_path,
            extra_mounts=list(self.revision.mounts.extra_mounts),
            # Revision fields - execution
            runtime_variant=self.revision.execution.runtime_variant,
            startup_command=self.revision.execution.startup_command,
            bootstrap_script=self.revision.execution.bootstrap_script,
            environ=dict(self.revision.execution.environ)
            if self.revision.execution.environ
            else {},
            callback_url=self.revision.execution.callback_url,
            # Default state fields
            lifecycle_stage=EndpointLifecycle.PENDING,
            retries=0,
        )


@dataclass
class EndpointLifecycleBatchUpdaterSpec(BatchUpdaterSpec[EndpointRow]):
    """BatchUpdaterSpec for batch updating endpoint lifecycle status.

    Used for transitioning multiple endpoints between lifecycle states.
    """

    lifecycle_stage: EndpointLifecycle

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        return {"lifecycle_stage": self.lifecycle_stage}



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.deployment.creator import ModelRevisionCreator
from ai.backend.manager.data.deployment.types import ModelRevisionData
from ai.backend.manager.services.deployment.actions.model_revision.base import (
    ModelRevisionBaseAction,
)


@dataclass
class CreateModelRevisionAction(ModelRevisionBaseAction):
    creator: ModelRevisionCreator

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateModelRevisionActionResult(BaseActionResult):
    revision: ModelRevisionData

    @override
    def entity_id(self) -> str | None:
        return str(self.revision.id)



from __future__ import annotations

from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.audit_log.actions.create import (
    CreateAuditLogAction,
    CreateAuditLogActionResult,
)
from ai.backend.manager.services.audit_log.actions.search import (
    SearchAuditLogsAction,
    SearchAuditLogsActionResult,
)
from ai.backend.manager.services.audit_log.service import AuditLogService


class AuditLogProcessors(AbstractProcessorPackage):
    create: ActionProcessor[CreateAuditLogAction, CreateAuditLogActionResult]
    search: ActionProcessor[SearchAuditLogsAction, SearchAuditLogsActionResult]

    def __init__(self, service: AuditLogService, action_monitors: list[ActionMonitor]) -> None:
        self.create = ActionProcessor(service.create, action_monitors)
        self.search = ActionProcessor(service.search, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateAuditLogAction.spec(),
            SearchAuditLogsAction.spec(),
        ]



from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.permission.role import (
    UserRoleAssignmentData,
    UserRoleAssignmentInput,
)
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class AssignRoleAction(RoleAction):
    input: UserRoleAssignmentInput

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.ROLE_ASSIGNMENT

    @override
    def entity_id(self) -> str | None:
        return str(self.input.user_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class AssignRoleActionResult(BaseActionResult):
    data: UserRoleAssignmentData

    @override
    def entity_id(self) -> str | None:
        return None



"""Resource Slot Normalization Row models.

Database models for normalized resource slot management:
- ResourceSlotTypeRow: Registry of known resource slot types and display metadata
- AgentResourceRow: Per-agent, per-slot resource capacity and usage
- ResourceAllocationRow: Per-kernel, per-slot resource allocation
"""

from __future__ import annotations

import uuid
from datetime import datetime
from decimal import Decimal

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column

from ai.backend.manager.models.base import (
    GUID,
    Base,
    PydanticColumn,
)
from ai.backend.manager.models.resource_slot.types import NumberFormat

__all__ = (
    "ResourceSlotTypeRow",
    "AgentResourceRow",
    "ResourceAllocationRow",
)


class ResourceSlotTypeRow(Base):  # type: ignore[misc]
    """Registry of known resource slot types with display metadata.

    Primary key is slot_name (e.g., 'cpu', 'mem', 'cuda.device').
    """

    __tablename__ = "resource_slot_types"

    slot_name: Mapped[str] = mapped_column("slot_name", sa.String(length=64), primary_key=True)
    slot_type: Mapped[str] = mapped_column("slot_type", sa.String(length=16), nullable=False)
    display_name: Mapped[str] = mapped_column(
        "display_name",
        sa.String(length=128),
        nullable=False,
        default="",
        server_default=sa.text("''"),
    )
    description: Mapped[str] = mapped_column(
        "description",
        sa.String(length=256),
        nullable=False,
        default="",
        server_default=sa.text("''"),
    )
    display_unit: Mapped[str] = mapped_column(
        "display_unit",
        sa.String(length=32),
        nullable=False,
        default="",
        server_default=sa.text("''"),
    )
    display_icon: Mapped[str] = mapped_column(
        "display_icon",
        sa.String(length=64),
        nullable=False,
        default="",
        server_default=sa.text("''"),
    )
    number_format: Mapped[NumberFormat] = mapped_column(
        "number_format",
        PydanticColumn(NumberFormat),
        nullable=False,
        default=NumberFormat(),
        server_default=sa.text(r"""'{"binary"\:false,"round_length"\:0}'::jsonb"""),
    )
    rank: Mapped[int] = mapped_column(
        "rank", sa.Integer, nullable=False, server_default=sa.text("0")
    )
    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )
    updated_at: Mapped[datetime] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        onupdate=sa.func.now(),
    )


class AgentResourceRow(Base):  # type: ignore[misc]
    """Per-agent, per-slot resource capacity and usage.

    Composite primary key: (agent_id, slot_name).
    """

    __tablename__ = "agent_resources"

    agent_id: Mapped[str] = mapped_column("agent_id", sa.String(length=64), primary_key=True)
    slot_name: Mapped[str] = mapped_column("slot_name", sa.String(length=64), primary_key=True)
    capacity: Mapped[Decimal] = mapped_column(
        "capacity", sa.Numeric(precision=24, scale=6), nullable=False
    )
    used: Mapped[Decimal] = mapped_column(
        "used", sa.Numeric(precision=24, scale=6), nullable=False, server_default=sa.text("0")
    )
    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )
    updated_at: Mapped[datetime] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        onupdate=sa.func.now(),
    )

    __table_args__ = (
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["agents.id"],
            name="fk_agent_resources_agent_id_agents",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["slot_name"],
            ["resource_slot_types.slot_name"],
            name="fk_agent_resources_slot_name_resource_slot_types",
        ),
        sa.Index("ix_agent_resources_slot_name", "slot_name"),
        sa.Index(
            "ix_agent_resources_agent_avail",
            "agent_id",
            "slot_name",
            "capacity",
            "used",
        ),
    )


class ResourceAllocationRow(Base):  # type: ignore[misc]
    """Per-kernel, per-slot resource allocation.

    Composite primary key: (kernel_id, slot_name).
    """

    __tablename__ = "resource_allocations"

    kernel_id: Mapped[uuid.UUID] = mapped_column("kernel_id", GUID, primary_key=True)
    slot_name: Mapped[str] = mapped_column("slot_name", sa.String(length=64), primary_key=True)
    requested: Mapped[Decimal] = mapped_column(
        "requested", sa.Numeric(precision=24, scale=6), nullable=False
    )
    used: Mapped[Decimal | None] = mapped_column(
        "used", sa.Numeric(precision=24, scale=6), nullable=True
    )
    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )
    used_at: Mapped[datetime | None] = mapped_column(
        "used_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )
    free_at: Mapped[datetime | None] = mapped_column(
        "free_at",
        sa.DateTime(timezone=True),
        nullable=True,
    )

    __table_args__ = (
        sa.ForeignKeyConstraint(
            ["kernel_id"],
            ["kernels.id"],
            name="fk_resource_allocations_kernel_id_kernels",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["slot_name"],
            ["resource_slot_types.slot_name"],
            name="fk_resource_allocations_slot_name_resource_slot_types",
        ),
        sa.Index("ix_resource_allocations_slot_name", "slot_name"),
        sa.Index("ix_ra_kernel_slot", "kernel_id", "slot_name"),
        sa.Index(
            "ix_ra_occupied",
            "kernel_id",
            "slot_name",
            postgresql_where=sa.text("free_at IS NULL"),
        ),
    )



"""Query conditions and orders for Resource Usage History repository."""

from __future__ import annotations

import uuid
from datetime import date, datetime

import sqlalchemy as sa

from ai.backend.manager.models.resource_usage_history import (
    DomainUsageBucketRow,
    KernelUsageRecordRow,
    ProjectUsageBucketRow,
    UserUsageBucketRow,
)
from ai.backend.manager.repositories.base.types import QueryCondition, QueryOrder


class KernelUsageRecordConditions:
    """Query conditions for KernelUsageRecordRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelUsageRecordRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_kernel_id(kernel_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelUsageRecordRow.kernel_id == kernel_id

        return inner

    @staticmethod
    def by_session_id(session_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelUsageRecordRow.session_id == session_id

        return inner

    @staticmethod
    def by_user_uuid(user_uuid: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelUsageRecordRow.user_uuid == user_uuid

        return inner

    @staticmethod
    def by_project_id(project_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelUsageRecordRow.project_id == project_id

        return inner

    @staticmethod
    def by_period_range(start: datetime, end: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                KernelUsageRecordRow.period_start >= start,
                KernelUsageRecordRow.period_end <= end,
            )

        return inner


class KernelUsageRecordOrders:
    """Query orders for KernelUsageRecordRow."""

    @staticmethod
    def by_period_start(ascending: bool = True) -> QueryOrder:
        col = KernelUsageRecordRow.period_start
        return col.asc() if ascending else col.desc()


class DomainUsageBucketConditions:
    """Query conditions for DomainUsageBucketRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_domain_name(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_resource_group_contains(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.resource_group.like(f"%{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_equals(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_resource_group_starts_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.resource_group.like(f"{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_ends_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.resource_group.like(f"%{resource_group}")

        return inner

    @staticmethod
    def by_domain_name_contains(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.domain_name.like(f"%{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_equals(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_domain_name_starts_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.domain_name.like(f"{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_ends_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.domain_name.like(f"%{domain_name}")

        return inner

    @staticmethod
    def by_period_range(start: date, end: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                DomainUsageBucketRow.period_start >= start,
                DomainUsageBucketRow.period_start <= end,
            )

        return inner

    @staticmethod
    def by_period_start(period_start: date) -> QueryCondition:
        """Filter by exact period start date."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_start == period_start

        return inner

    @staticmethod
    def by_period_start_not_equals(period_start: date) -> QueryCondition:
        """Filter by period start date not equal to the given value."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_start != period_start

        return inner

    @staticmethod
    def by_period_start_before(before: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_start < before

        return inner

    @staticmethod
    def by_period_start_after(after: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_start > after

        return inner

    @staticmethod
    def by_period_end(period_end: date) -> QueryCondition:
        """Filter by exact period end date."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_end == period_end

        return inner

    @staticmethod
    def by_period_end_not_equals(period_end: date) -> QueryCondition:
        """Filter by period end date not equal to the given value."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_end != period_end

        return inner

    @staticmethod
    def by_period_end_before(before: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_end < before

        return inner

    @staticmethod
    def by_period_end_after(after: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return DomainUsageBucketRow.period_end > after

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get period_start of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DomainUsageBucketRow.period_start)
                .where(DomainUsageBucketRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DomainUsageBucketRow.period_start < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get period_start of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(DomainUsageBucketRow.period_start)
                .where(DomainUsageBucketRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return DomainUsageBucketRow.period_start > subquery

        return inner


class DomainUsageBucketOrders:
    """Query orders for DomainUsageBucketRow."""

    @staticmethod
    def by_period_start(ascending: bool = True) -> QueryOrder:
        col = DomainUsageBucketRow.period_start
        return col.asc() if ascending else col.desc()


class ProjectUsageBucketConditions:
    """Query conditions for ProjectUsageBucketRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_domain_name(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_resource_group_contains(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.resource_group.like(f"%{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_equals(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_resource_group_starts_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.resource_group.like(f"{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_ends_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.resource_group.like(f"%{resource_group}")

        return inner

    @staticmethod
    def by_domain_name_contains(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.domain_name.like(f"%{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_equals(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_domain_name_starts_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.domain_name.like(f"{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_ends_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.domain_name.like(f"%{domain_name}")

        return inner

    @staticmethod
    def by_project_id(project_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.project_id == project_id

        return inner

    @staticmethod
    def by_period_range(start: date, end: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                ProjectUsageBucketRow.period_start >= start,
                ProjectUsageBucketRow.period_start <= end,
            )

        return inner

    @staticmethod
    def by_period_start(period_start: date) -> QueryCondition:
        """Filter by exact period start date."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_start == period_start

        return inner

    @staticmethod
    def by_period_start_not_equals(period_start: date) -> QueryCondition:
        """Filter by period start date not equal to the given value."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_start != period_start

        return inner

    @staticmethod
    def by_period_start_before(before: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_start < before

        return inner

    @staticmethod
    def by_period_start_after(after: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_start > after

        return inner

    @staticmethod
    def by_period_end(period_end: date) -> QueryCondition:
        """Filter by exact period end date."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_end == period_end

        return inner

    @staticmethod
    def by_period_end_not_equals(period_end: date) -> QueryCondition:
        """Filter by period end date not equal to the given value."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_end != period_end

        return inner

    @staticmethod
    def by_period_end_before(before: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_end < before

        return inner

    @staticmethod
    def by_period_end_after(after: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ProjectUsageBucketRow.period_end > after

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get period_start of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(ProjectUsageBucketRow.period_start)
                .where(ProjectUsageBucketRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return ProjectUsageBucketRow.period_start < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get period_start of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(ProjectUsageBucketRow.period_start)
                .where(ProjectUsageBucketRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return ProjectUsageBucketRow.period_start > subquery

        return inner


class ProjectUsageBucketOrders:
    """Query orders for ProjectUsageBucketRow."""

    @staticmethod
    def by_period_start(ascending: bool = True) -> QueryOrder:
        col = ProjectUsageBucketRow.period_start
        return col.asc() if ascending else col.desc()


class UserUsageBucketConditions:
    """Query conditions for UserUsageBucketRow."""

    @staticmethod
    def by_resource_group(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_domain_name(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_resource_group_contains(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.resource_group.like(f"%{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_equals(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.resource_group == resource_group

        return inner

    @staticmethod
    def by_resource_group_starts_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.resource_group.like(f"{resource_group}%")

        return inner

    @staticmethod
    def by_resource_group_ends_with(resource_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.resource_group.like(f"%{resource_group}")

        return inner

    @staticmethod
    def by_domain_name_contains(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.domain_name.like(f"%{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_equals(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.domain_name == domain_name

        return inner

    @staticmethod
    def by_domain_name_starts_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.domain_name.like(f"{domain_name}%")

        return inner

    @staticmethod
    def by_domain_name_ends_with(domain_name: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.domain_name.like(f"%{domain_name}")

        return inner

    @staticmethod
    def by_user_uuid(user_uuid: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.user_uuid == user_uuid

        return inner

    @staticmethod
    def by_project_id(project_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.project_id == project_id

        return inner

    @staticmethod
    def by_period_range(start: date, end: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return sa.and_(
                UserUsageBucketRow.period_start >= start,
                UserUsageBucketRow.period_start <= end,
            )

        return inner

    @staticmethod
    def by_period_start(period_start: date) -> QueryCondition:
        """Filter by exact period start date."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_start == period_start

        return inner

    @staticmethod
    def by_period_start_not_equals(period_start: date) -> QueryCondition:
        """Filter by period start date not equal to the given value."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_start != period_start

        return inner

    @staticmethod
    def by_period_start_before(before: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_start < before

        return inner

    @staticmethod
    def by_period_start_after(after: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_start > after

        return inner

    @staticmethod
    def by_period_end(period_end: date) -> QueryCondition:
        """Filter by exact period end date."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_end == period_end

        return inner

    @staticmethod
    def by_period_end_not_equals(period_end: date) -> QueryCondition:
        """Filter by period end date not equal to the given value."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_end != period_end

        return inner

    @staticmethod
    def by_period_end_before(before: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_end < before

        return inner

    @staticmethod
    def by_period_end_after(after: date) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserUsageBucketRow.period_end > after

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get period_start of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(UserUsageBucketRow.period_start)
                .where(UserUsageBucketRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return UserUsageBucketRow.period_start < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get period_start of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(UserUsageBucketRow.period_start)
                .where(UserUsageBucketRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return UserUsageBucketRow.period_start > subquery

        return inner


class UserUsageBucketOrders:
    """Query orders for UserUsageBucketRow."""

    @staticmethod
    def by_period_start(ascending: bool = True) -> QueryOrder:
        col = UserUsageBucketRow.period_start
        return col.asc() if ascending else col.desc()



from ai.backend.common.types import MountPermission, QuotaScopeID, VFolderID, VFolderUsageMode
from ai.backend.manager.data.vfolder.types import (
    VFolderInvitationState,
    VFolderOperationStatus,
    VFolderOwnershipType,
)
from ai.backend.manager.data.vfolder.types import VFolderMountPermission as VFolderPermission

from .row import (
    DEAD_VFOLDER_STATUSES,
    HARD_DELETED_VFOLDER_STATUSES,
    SOFT_DELETED_VFOLDER_STATUSES,
    VFolderCloneInfo,
    VFolderDeletionInfo,
    VFolderInvitationRow,
    VFolderPermissionRow,
    VFolderPermissionSetAlias,
    VFolderPermissionValidator,
    VFolderRow,
    VFolderStatusSet,
    delete_vfolder_relation_rows,
    ensure_host_permission_allowed,
    ensure_quota_scope_accessible_by_user,
    filter_host_allowed_permission,
    get_allowed_vfolder_hosts_by_group,
    get_allowed_vfolder_hosts_by_user,
    get_sessions_by_mounted_folder,
    initiate_vfolder_deletion,
    is_mount_duplicate,
    is_unmanaged,
    prepare_vfolder_mounts,
    query_accessible_vfolders,
    update_vfolder_status,
    verify_vfolder_name,
    vfolder_invitations,
    vfolder_permissions,
    vfolder_status_map,
    vfolders,
)
from .row import (
    get_permission_ctx as get_permission_ctx,
)

__all__ = (
    "DEAD_VFOLDER_STATUSES",
    "HARD_DELETED_VFOLDER_STATUSES",
    "SOFT_DELETED_VFOLDER_STATUSES",
    "MountPermission",
    "QuotaScopeID",
    "VFolderCloneInfo",
    "VFolderDeletionInfo",
    "VFolderID",
    "VFolderInvitationRow",
    "VFolderInvitationState",
    "VFolderOperationStatus",
    "VFolderOwnershipType",
    "VFolderPermission",
    "VFolderPermissionRow",
    "VFolderPermissionSetAlias",
    "VFolderPermissionValidator",
    "VFolderRow",
    "VFolderStatusSet",
    "VFolderUsageMode",
    "delete_vfolder_relation_rows",
    "ensure_host_permission_allowed",
    "ensure_quota_scope_accessible_by_user",
    "filter_host_allowed_permission",
    "get_allowed_vfolder_hosts_by_group",
    "get_allowed_vfolder_hosts_by_user",
    "get_sessions_by_mounted_folder",
    "initiate_vfolder_deletion",
    "is_mount_duplicate",
    "is_unmanaged",
    "prepare_vfolder_mounts",
    "query_accessible_vfolders",
    "update_vfolder_status",
    "verify_vfolder_name",
    "vfolder_invitations",
    "vfolder_permissions",
    "vfolder_status_map",
    "vfolders",
)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from ai.backend.manager.repositories.types import RepositoryArgs

from .repository import AuditLogRepository


@dataclass
class AuditLogRepositories:
    repository: AuditLogRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        return cls(
            repository=AuditLogRepository(db=args.db),
        )



"""Add app_configs table for frontend configuration

Revision ID: d811b103dbfc
Revises: 09206ac04fd3
Create Date: 2025-10-26 15:11:41.781361

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "d811b103dbfc"
down_revision = "09206ac04fd3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "app_configs",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column(
            "scope_type",
            sa.Enum("DOMAIN", "PROJECT", "USER", name="app_config_scope_type"),
            nullable=False,
        ),
        sa.Column("scope_id", sa.String(length=256), nullable=False),
        sa.Column("extra_config", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column(
            "modified_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_app_configs")),
        sa.UniqueConstraint("scope_type", "scope_id", name="uq_app_configs_scope"),
    )
    op.create_index(op.f("ix_app_configs_scope_id"), "app_configs", ["scope_id"], unique=False)
    op.create_index(op.f("ix_app_configs_scope_type"), "app_configs", ["scope_type"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_app_configs_scope_type"), table_name="app_configs")
    op.drop_index(op.f("ix_app_configs_scope_id"), table_name="app_configs")
    op.drop_table("app_configs")
    op.execute("DROP TYPE IF EXISTS app_config_scope_type")
    # ### end Alembic commands ###
